"""Build SDMX-REST data queries."""

from abc import abstractmethod
from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Optional, Sequence, Union

import msgspec

from pysdmx.api.dc.query import (
    LogicalOperator,
    MultiFilter,
    NumberFilter,
    Operator,
    TextFilter,
)
from pysdmx.api.qb.util import (
    ApiVersion,
    check_multiple_data_context,
    REST_ALL,
)
from pysdmx.errors import Invalid


class DataContext(Enum):
    """The context of the data query."""

    DATA_STRUCTURE = "datastructure"
    DATAFLOW = "dataflow"
    PROVISION_AGREEMENT = "provisionagreement"
    ALL = REST_ALL


class DataFormat(Enum):
    """The response formats for data queries."""

    SDMX_JSON_1_0_0 = "application/vnd.sdmx.data+json;version=1.0.0"
    SDMX_JSON_2_0_0 = "application/vnd.sdmx.data+json;version=2.0.0"
    SDMX_CSV_1_0_0 = "application/vnd.sdmx.data+csv;version=1.0.0"
    SDMX_CSV_2_0_0 = "application/vnd.sdmx.data+csv;version=2.0.0"
    SDMX_ML_2_1_GEN = "application/vnd.sdmx.genericdata+xml;version=2.1"
    SDMX_ML_2_1_STR = (
        "application/vnd.sdmx.structurespecificdata+xml;version=2.1"
    )
    SDMX_ML_2_1_GENTS = (
        "application/vnd.sdmx.generictimeseriesdata+xml;version=2.1"
    )
    SDMX_ML_2_1_STRTS = (
        "application/vnd.sdmx.structurespecifictimeseriesdata+xml;version=2.1"
    )
    SDMX_ML_3_0 = "application/vnd.sdmx.data+xml;version=3.0.0"


class _CoreDataQuery(msgspec.Struct, frozen=True, omit_defaults=True):

    def get_url(self, version: ApiVersion, omit_defaults: bool = False) -> str:
        """The URL for the query in the selected SDMX-REST API version."""
        self._validate_query(version)
        if omit_defaults:
            return self._create_short_query(version)
        else:
            return self._create_full_query(version)

    def validate(self) -> None:
        """Validate the query."""
        try:
            self._get_decoder().decode(_encoder.encode(self))
        except msgspec.DecodeError as err:
            raise Invalid("Invalid Schema Query", str(err)) from err

    @abstractmethod
    def _get_decoder(self) -> msgspec.json.Decoder:  # type: ignore[type-arg]
        """Returns the decoder to be used for validation."""

    def _validate_context(
        self, context: DataContext, api_version: ApiVersion
    ) -> None:
        if api_version < ApiVersion.V2_0_0 and context in [
            DataContext.DATA_STRUCTURE,
            DataContext.PROVISION_AGREEMENT,
        ]:
            raise Invalid(
                "Validation Error",
                f"{context} is not valid for SDMX-REST {api_version.value}.",
            )

    def _check_multiple_contexts(
        self,
        agency_id: Union[str, Sequence[str]],
        resource_id: Union[str, Sequence[str]],
        version: Union[str, Sequence[str]],
        key: Union[str, Sequence[str]],
        api_version: ApiVersion,
    ) -> None:
        check_multiple_data_context("agency", agency_id, api_version)
        check_multiple_data_context("resource", resource_id, api_version)
        check_multiple_data_context("version", version, api_version)
        check_multiple_data_context("key", key, api_version)

    def _check_resource_id(
        self, resource_id: Union[str, Sequence[str]], api_version: ApiVersion
    ) -> None:
        if api_version < ApiVersion.V2_0_0 and resource_id == REST_ALL:
            raise Invalid(
                "Validation Error",
                (
                    f"A dataflow must be provided in SDMX-REST "
                    f"{api_version.value}."
                ),
            )

    def _check_components(
        self,
        components: Union[MultiFilter, None, NumberFilter, TextFilter],
        api_version: ApiVersion,
    ) -> None:
        if api_version < ApiVersion.V2_0_0 and components:
            raise Invalid(
                "Validation Error",
                (
                    "Components filter is not supported in "
                    f"SDMX-REST {api_version.value}."
                ),
            )

    def _to_kw(self, val: str, ver: ApiVersion) -> str:
        if val == "*" and ver < ApiVersion.V2_0_0:
            val = "all"
        elif val == "~" and ver < ApiVersion.V2_0_0:
            val = "latest"
        return val

    def _to_kws(self, vals: Union[str, Sequence[str]], ver: ApiVersion) -> str:
        vals = [vals] if isinstance(vals, str) else vals
        mapped = [self._to_kw(v, ver) for v in vals]
        return ",".join(mapped)

    def _get_v2_context_id(
        self,
        context: DataContext,
        agency_id: Union[str, Sequence[str]],
        resource_id: Union[str, Sequence[str]],
        version: Union[str, Sequence[str]],
        api_version: ApiVersion,
    ) -> str:
        o = f"/{context.value}"
        a = self._to_kws(agency_id, api_version)
        r = self._to_kws(resource_id, api_version)
        v = self._to_kws(version, api_version)
        o += f"/{a}/{r}/{v}"
        return o

    def _get_v1_context_id(
        self,
        agency_id: str,
        resource_id: str,
        version: str,
        api_version: ApiVersion,
    ) -> str:
        a = self._to_kw(agency_id, api_version)
        r = self._to_kw(resource_id, api_version)
        v = (
            self._to_kw(version, api_version)
            if version != REST_ALL
            else "latest"
        )
        return f"/{a},{r},{v}"

    def _get_short_v2_path(
        self,
        resource: str,
        context: DataContext,
        agency_id: Union[str, Sequence[str]],
        resource_id: Union[str, Sequence[str]],
        version: Union[str, Sequence[str]],
        key: Union[str, Sequence[str]],
        api_version: ApiVersion,
        component_id: Union[str, Sequence[str], None] = None,
    ) -> str:
        d = (
            f"/{self._to_kws(component_id, api_version)}"
            if component_id and component_id != REST_ALL
            else ""
        )
        k = (
            f"/{self._to_kws(key, api_version)}{d}"
            if d or key != REST_ALL
            else ""
        )
        v = (
            f"/{self._to_kws(version, api_version)}{k}"
            if k or version != REST_ALL
            else ""
        )
        r = (
            f"/{self._to_kws(resource_id, api_version)}{v}"
            if v or resource_id != REST_ALL
            else ""
        )
        a = (
            f"/{self._to_kws(agency_id, api_version)}{r}"
            if r or agency_id != REST_ALL
            else ""
        )
        c = f"/{context.value}{a}" if a or context != DataContext.ALL else ""
        return f"/{resource}{c}"

    def _get_short_v1_path(
        self,
        resource: str,
        agency_id: str,
        resource_id: str,
        version: str,
        key: Union[str, Sequence[str]],
        api_version: ApiVersion,
    ) -> str:
        v = (
            f",{self._to_kw(version, api_version)}"
            if version != REST_ALL
            else ""
        )
        kr = self._to_kw(resource_id, api_version)
        r = f"{kr}{v}"
        if agency_id != REST_ALL or version != REST_ALL:
            ka = self._to_kw(agency_id, api_version)
            a = f"{ka},{r}"
        else:
            a = f"{r}"
        k = f"/{key}" if key != REST_ALL else ""
        return f"/{resource}/{a}{k}"

    def _append_qs_param(
        self, qs: str, value: Any, field: str, disp_value: Any = None
    ) -> str:
        if value:
            if qs:
                qs += "&"
            qs += f"{field}={disp_value if disp_value else value}"
        return qs

    def _create_component_filters(
        self, components: Union[MultiFilter, NumberFilter, TextFilter]
    ) -> str:
        if isinstance(components, MultiFilter):
            if components.operator == LogicalOperator.OR:
                raise Invalid(
                    "Validation Error",
                    "OR operator is not supported for MultiFilter.",
                )
            flts_by_comp = defaultdict(list)
            for f in components.filters:
                if isinstance(f, (NumberFilter, TextFilter)):
                    flts_by_comp[f.field].append(f)
                else:
                    raise Invalid(
                        "Validation Error",
                        f"Unsupported filter type: {f}.",
                    )
            flts = []
            for k, v in flts_by_comp.items():
                if len(v) > 1:
                    flts.append(_create_component_mult_filter(k, v))
                else:
                    flts.append(_create_component_filter(v[0]))
            return "&".join(flts)
        else:
            return _create_component_filter(components)

    @abstractmethod
    def _validate_query(self, version: ApiVersion) -> None:
        """Any additional validation steps to be performed by subclasses."""

    @abstractmethod
    def _create_full_query(self, ver: ApiVersion) -> str:
        """Creates a URL, with default values."""

    @abstractmethod
    def _create_short_query(self, ver: ApiVersion) -> str:
        """Creates a URL, omitting default values when possible."""


class DataQuery(_CoreDataQuery, frozen=True, omit_defaults=True):
    """A data query.

    Data queries allow retrieving statistical data.

    Attributes:
        context: The context for which data must be retrieved.
        agency_id: The agency maintaining the context to be considered.
        resource_id: The id of the context to be considered.
        version: The version of the context to be considered.
        key: The combination of dimension values identifying the slice
            of the cube for which data should be returned. Wildcarding
            is supported via the * operator.
        components: The component values to be used for data filtering.
        updated_after: The last time the query was performed by the client.
        first_n_obs: The maximum number of observations to be returned for
            each of the matching series, starting from the first observation.
        last_n_obs: The maximum number of observations to be returned for
            each of the matching series, counting back from the most recent
            observation.
        obs_dim: The ID of the dimension to be attached at the observation
            level.
        attributes: The attributes to be returned. Possible options are:
            `dsd` (all the attributes defined in the DSD), `msd` (all the
            reference metadata attributes), `dataset` (all the attributes
            attached to the dataset-level), `series` (all the attributes
            attached to the series-level), `obs` (all the attributes
            attached to the observation-level), `all` (all attributes),
            `none` (no attributes), {attribute_id}: The ID of the one or
            more attributes to be returned.
        measures: The measures to be returned. Possible options are:
            `all` (all measures), `none` (no measure), {measure_id}:
            The ID of the one or more measures to be returned.
        include_history: Retrieve previous versions of the data, as they
            were disseminated in the past.
    """

    context: DataContext = DataContext.ALL
    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_ALL
    key: Union[str, Sequence[str]] = REST_ALL
    components: Union[MultiFilter, None, NumberFilter, TextFilter] = None
    updated_after: Optional[datetime] = None
    first_n_obs: Optional[Annotated[int, msgspec.Meta(gt=0)]] = None
    last_n_obs: Optional[Annotated[int, msgspec.Meta(gt=0)]] = None
    obs_dimension: Optional[str] = None
    attributes: Union[
        str,
        Sequence[str],
    ] = "dsd"
    measures: Union[
        str,
        Sequence[str],
    ] = "all"
    include_history: bool = False

    def _validate_query(self, api_version: ApiVersion) -> None:
        self.validate()
        super()._validate_context(self.context, api_version)
        super()._check_multiple_contexts(
            self.agency_id,
            self.resource_id,
            self.version,
            self.key,
            api_version,
        )
        super()._check_resource_id(self.resource_id, api_version)
        super()._check_components(self.components, api_version)

    def _get_decoder(self) -> msgspec.json.Decoder:  # type: ignore[type-arg]
        return _data_decoder

    def __get_short_v2_qs(self, api_version: ApiVersion) -> str:
        qs = ""
        if self.updated_after:
            qs = super()._append_qs_param(
                qs,
                self.updated_after,
                "updatedAfter",
                self.updated_after.isoformat("T", "seconds"),
            )
        qs = super()._append_qs_param(
            qs, self.first_n_obs, "firstNObservations"
        )
        qs = super()._append_qs_param(qs, self.last_n_obs, "lastNObservations")
        qs = super()._append_qs_param(
            qs, self.obs_dimension, "dimensionAtObservation"
        )
        if self.attributes != "dsd":
            qs = super()._append_qs_param(
                qs, super()._to_kws(self.attributes, api_version), "attributes"
            )
        if self.measures != "all":
            qs = super()._append_qs_param(
                qs, super()._to_kws(self.measures, api_version), "measures"
            )
        qs = super()._append_qs_param(
            qs,
            self.include_history,
            "includeHistory",
            str(self.include_history).lower(),
        )
        return f"?{qs}" if qs else qs

    def __get_short_v1_qs(self, api_version: ApiVersion) -> str:
        qs = ""
        if self.updated_after:
            qs = super()._append_qs_param(
                qs,
                self.updated_after,
                "updatedAfter",
                self.updated_after.isoformat("T", "seconds"),
            )
        qs = super()._append_qs_param(
            qs, self.first_n_obs, "firstNObservations"
        )
        qs = super()._append_qs_param(qs, self.last_n_obs, "lastNObservations")
        qs = super()._append_qs_param(
            qs, self.obs_dimension, "dimensionAtObservation"
        )
        detail = self.__get_v1_detail(api_version)
        if detail != "full":
            qs = super()._append_qs_param(qs, detail, "detail")
        qs = super()._append_qs_param(
            qs,
            self.include_history,
            "includeHistory",
            str(self.include_history).lower(),
        )
        return f"?{qs}" if qs else qs

    def __get_v1_detail(self, api_version: ApiVersion) -> str:
        if self.measures in ["OBS_VALUE", "all"] and self.attributes == "dsd":
            return "full"
        elif (
            self.measures in ["OBS_VALUE", "all"] and self.attributes == "none"
        ):
            return "dataonly"
        if self.measures == "none" and self.attributes == "series":
            return "serieskeysonly"
        if self.measures == "none" and self.attributes == "dsd":
            return "nodata"
        else:
            raise Invalid(
                "Validation Error",
                (
                    f"{self.attributes} and {self.measures} is not a valid "
                    "combination for the detail attribute in SDMX-REST "
                    f"{api_version.value}."
                ),
            )

    def _create_full_query(self, api_version: ApiVersion) -> str:
        o = "/data"
        if api_version >= ApiVersion.V2_0_0:
            c = super()._get_v2_context_id(
                self.context,
                self.agency_id,
                self.resource_id,
                self.version,
                api_version,
            )
        else:
            c = super()._get_v1_context_id(
                self.agency_id,  # type: ignore[arg-type]
                self.resource_id,  # type: ignore[arg-type]
                self.version,  # type: ignore[arg-type]
                api_version,
            )
        o += f"{c}/{super()._to_kws(self.key, api_version)}"
        qs = ""
        if self.components:
            qs += self._create_component_filters(self.components)
        if self.updated_after:
            qs = super()._append_qs_param(
                qs,
                self.updated_after,
                "updatedAfter",
                self.updated_after.isoformat("T", "seconds"),
            )
        qs = super()._append_qs_param(
            qs, self.first_n_obs, "firstNObservations"
        )
        qs = super()._append_qs_param(qs, self.last_n_obs, "lastNObservations")
        qs = super()._append_qs_param(
            qs, self.obs_dimension, "dimensionAtObservation"
        )
        if api_version >= ApiVersion.V2_0_0:
            qs = super()._append_qs_param(
                qs,
                self.attributes,
                "attributes",
                super()._to_kws(self.attributes, api_version),
            )
            qs = super()._append_qs_param(
                qs,
                self.measures,
                "measures",
                super()._to_kws(self.measures, api_version),
            )
        else:
            qs = super()._append_qs_param(
                qs, self.__get_v1_detail(api_version), "detail"
            )
        qs = super()._append_qs_param(
            qs, str(self.include_history).lower(), "includeHistory"
        )
        return f"{o}?{qs}"

    def _create_short_query(self, api_version: ApiVersion) -> str:
        if api_version >= ApiVersion.V2_0_0:
            p = super()._get_short_v2_path(
                "data",
                self.context,
                self.agency_id,
                self.resource_id,
                self.version,
                self.key,
                api_version,
            )
            q = self.__get_short_v2_qs(api_version)
            o = f"{p}{q}"
        else:
            p = super()._get_short_v1_path(
                "data",
                self.agency_id,  # type: ignore[arg-type]
                self.resource_id,  # type: ignore[arg-type]
                self.version,  # type: ignore[arg-type]
                self.key,
                api_version,
            )
            q = self.__get_short_v1_qs(api_version)
            o = f"{p}{q}"
        return o


def __map_like_operator(value: Any) -> str:
    if isinstance(value, str):
        if value.startswith("%") and value.endswith("%"):
            out = "co"
        elif value.startswith("%"):
            out = "ew"
        elif value.endswith("%"):
            out = "sw"
        else:
            raise Invalid(
                "Validation Error",
                (
                    f"The supplied value ({value}) is not correct "
                    "for the LIKE operator."
                ),
            )
        return out
    else:
        raise Invalid(
            "Validation Error",
            (
                f"The supplied value ({value}) for the LIKE operator "
                "should be a string."
            ),
        )


def __map_operator(op: Operator, value: Any) -> str:
    out = ""
    if op == Operator.NOT_EQUALS:
        out = "ne"
    elif op == Operator.LESS_THAN:
        out = "lt"
    elif op == Operator.LESS_THAN_OR_EQUAL:
        out = "le"
    elif op == Operator.GREATER_THAN:
        out = "gt"
    elif op == Operator.GREATER_THAN_OR_EQUAL:
        out = "ge"
    elif op == Operator.LIKE:
        out = __map_like_operator(value)
    elif op == Operator.NOT_LIKE:
        out = "nc"
    return out


def _create_component_filter(flt: Union[NumberFilter, TextFilter]) -> str:
    fld = flt.field
    val = flt.value
    if flt.operator in [Operator.LIKE, Operator.NOT_LIKE]:
        op = __map_operator(flt.operator, val)
        val = str(val).replace("%", "")
        return f"c[{fld}]={op}:{val}"
    elif flt.operator == Operator.IN:
        return f"c[{fld}]={','.join(val)}"  # type: ignore[arg-type]
    elif flt.operator == Operator.BETWEEN:
        return f"c[{fld}]=ge:{val[0]}+le:{val[1]}"  # type: ignore[index]
    elif flt.operator in [
        Operator.EQUALS,
        Operator.NOT_EQUALS,
        Operator.LIKE,
        Operator.NOT_LIKE,
        Operator.LESS_THAN,
        Operator.LESS_THAN_OR_EQUAL,
        Operator.GREATER_THAN,
        Operator.GREATER_THAN_OR_EQUAL,
    ]:
        op = __map_operator(flt.operator, val)
        if op:
            op += ":"
        return f"c[{fld}]={op}{val}"
    else:
        raise Invalid(
            "Validation Error",
            (f"The operator ({flt.operator}) is not supported."),
        )


def _create_component_mult_filter(
    comp: str, flts: Sequence[Union[NumberFilter, TextFilter]]
) -> str:
    cstr = [_create_component_filter(f) for f in flts]
    fstr = [s.replace(f"c[{comp}]=", "") for s in cstr]
    return f"c[{comp}]={'+'.join(fstr)}"


_data_decoder = msgspec.json.Decoder(DataQuery)
_encoder = msgspec.json.Encoder()



"""Collection of Fusion-JSON schemas for organisations."""

from collections import defaultdict
from typing import Dict, Optional, Sequence, Set

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.model import Agency, Contact, DataflowRef, DataProvider
from pysdmx.util import parse_urn


class FusionContact(Struct, frozen=True):
    """Fusion-JSON payload for a contact."""

    id: Optional[str] = None
    names: Optional[Sequence[FusionString]] = None
    departments: Optional[Sequence[FusionString]] = None
    roles: Optional[Sequence[FusionString]] = None
    telephone: Optional[Sequence[str]] = None
    fax: Optional[Sequence[str]] = None
    uri: Optional[Sequence[str]] = None
    email: Optional[Sequence[str]] = None

    def to_model(self) -> Contact:
        """Converts a FusionContact to a standard Contact."""
        n = self.names[0].value if self.names else None
        d = self.departments[0].value if self.departments else None
        r = self.roles[0].value if self.roles else None

        return Contact(
            self.id,
            n,
            d,
            r,
            self.telephone,
            self.fax,
            self.uri,
            self.email,
        )


class FusionAgency(Struct, frozen=True):
    """Fusion-JSON payload for an organisation."""

    id: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    contacts: Sequence[FusionContact] = ()

    def to_model(self, owner: Optional[str] = None) -> Agency:
        """Converts a FusionOrg to a standard Organisation."""
        d = self.descriptions[0].value if self.descriptions else None
        c = [c.to_model() for c in self.contacts]
        oid = f"{owner}.{self.id}" if owner and owner != "SDMX" else self.id
        if c:
            return Agency(
                id=oid, name=self.names[0].value, description=d, contacts=c
            )
        else:
            return Agency(id=oid, name=self.names[0].value, description=d)


class FusionProvider(Struct, frozen=True):
    """Fusion-JSON payload for an organisation."""

    id: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    contacts: Sequence[FusionContact] = ()

    def to_model(self, owner: Optional[str] = None) -> DataProvider:
        """Converts a FusionOrg to a standard Organisation."""
        d = self.descriptions[0].value if self.descriptions else None
        c = [c.to_model() for c in self.contacts]
        oid = f"{owner}.{self.id}" if owner and owner != "SDMX" else self.id
        if c:
            return DataProvider(
                id=oid, name=self.names[0].value, description=d, contacts=c
            )
        else:
            return DataProvider(
                id=oid, name=self.names[0].value, description=d
            )


class FusionAgencyScheme(Struct, frozen=True):
    """Fusion-JSON payload for an agency scheme."""

    agencyId: str
    items: Sequence[FusionAgency]

    def to_model(self) -> Sequence[Agency]:
        """Converts a FusionAgencyScheme to a list of Organisations."""
        return [o.to_model(self.agencyId) for o in self.items]


class FusionAgencyMessage(Struct, frozen=True):
    """Fusion-JSON payload for /agencyscheme queries."""

    AgencyScheme: Sequence[FusionAgencyScheme]

    def to_model(self) -> Sequence[Agency]:
        """Returns the requested list of agencies."""
        return self.AgencyScheme[0].to_model()


class FusionProvisionAgreement(Struct, frozen=True):
    """Fusion-JSON payload for a provision agreement."""

    structureUsage: str
    dataproviderRef: str


class FusionProviderScheme(Struct, frozen=True):
    """Fusion-JSON payload for a data provider scheme."""

    items: Sequence[FusionProvider]

    def __get_df_ref(self, ref: str) -> DataflowRef:
        a = parse_urn(ref)
        return DataflowRef(id=a.id, agency=a.agency, version=a.version)

    def to_model(
        self, pas: Sequence[FusionProvisionAgreement]
    ) -> Sequence[DataProvider]:
        """Converts a FusionProviderScheme to a list of Organisations."""
        if pas:
            paprs: Dict[str, Set[DataflowRef]] = defaultdict(set)
            for pa in pas:
                df = self.__get_df_ref(pa.structureUsage)
                pr = pa.dataproviderRef[pa.dataproviderRef.rindex(".") + 1 :]
                paprs[pr].add(df)
            prvs = [o.to_model() for o in self.items]
            return [
                DataProvider(
                    id=p.id,
                    name=p.name,
                    description=p.description,
                    contacts=p.contacts,
                    dataflows=list(paprs[p.id]),
                )
                for p in prvs
            ]
        else:
            return [o.to_model() for o in self.items]


class FusionProviderMessage(Struct, frozen=True):
    """Fusion-JSON payload for /dataproviderscheme queries."""

    DataProviderScheme: Sequence[FusionProviderScheme]
    ProvisionAgreement: Sequence[FusionProvisionAgreement] = ()

    def to_model(self) -> Sequence[DataProvider]:
        """Returns the requested list of providers."""
        return self.DataProviderScheme[0].to_model(self.ProvisionAgreement)



from datetime import timezone
from typing import Sequence, Union

import dateutil
import dateutil.parser
from parsy import digit, regex, string  # type: ignore[import-untyped]

from pysdmx.api.dc.query._model import (
    DateTimeFilter,
    Filter,
    LogicalOperator,
    MultiFilter,
    NumberFilter,
    TextFilter,
)
from pysdmx.api.dc.query._parsing_model import (
    _DateTime,
    _Field,
    _Filter,
    _Number,
    _String,
)


def __map_filter(f: _Filter) -> Filter:
    if isinstance(f.value, list):
        if isinstance(f.value[0], _Number):
            values = [v.value for v in f.value]
            return NumberFilter(f.field.name, f.operator, values)
        elif isinstance(f.value[0], _DateTime):
            values = [v.value for v in f.value]
            return DateTimeFilter(f.field.name, f.operator, values)
        else:
            values = [v.value for v in f.value]
            return TextFilter(f.field.name, f.operator, values)
    elif isinstance(f.value, _Number):
        return NumberFilter(f.field.name, f.operator, f.value.value)
    elif isinstance(f.value, _DateTime):
        return DateTimeFilter(f.field.name, f.operator, f.value.value)
    else:
        return TextFilter(
            f.field.name,
            f.operator,
            f.value.value,  # type: ignore[arg-type,union-attr]
        )


def __map_string(input: str) -> Union[_DateTime, _String]:
    rec = input[1:-1]
    try:
        dt = dateutil.parser.parse(rec)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return _DateTime(dt)
    except dateutil.parser.ParserError:
        return _String(rec)


def _to_response(filters: Sequence[_Filter]) -> Filter:
    if len(filters) == 1:
        return __map_filter(filters[0])
    else:
        flts = [__map_filter(f) for f in filters]
        return MultiFilter(flts, LogicalOperator.AND)


_pad = regex(r"\s*")  # optional whitespace
_field = regex(r"[A-Za-z0-9_@$\-]+").map(_Field)
_int_val = digit.at_least(1).concat().map(int).map(_Number)
_float_val = (
    (digit.many() + string(".").result(["."]) + digit.many())
    .concat()
    .map(float)
    .map(_Number)
)
_string_val = regex(r"'[^']*'").map(__map_string)
_value = _float_val | _int_val | _string_val
_in_vals = _value.sep_by(_pad + string(",") + _pad, min=1)



"""Message module.

This module contains the enumeration for the different types of messages that
can be written. It also contains the Header and Message classes that are used
to create the SDMX messages.

Classes:
    Header: Header for the SDMX messages.
    ActionType: Enumeration for the different types of actions that can be
        written.
    Message: Class that holds the content of the SDMX message.
    SubmissionResult: Class that represents the result of a submission.
"""

from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, Optional
import uuid

from msgspec import Struct

from pysdmx.errors import Invalid, NotFound
from pysdmx.model import Codelist, ConceptScheme
from pysdmx.model.__base import ItemScheme


class ActionType(Enum):
    """ActionType enumeration.

    Enumeration that withholds the Action type for writing purposes.
    """

    Append = "append"
    Replace = "replace"
    Delete = "delete"
    Information = "information"


class Header(Struct, frozen=True, kw_only=True):
    """Header for the SDMX messages."""

    id: str = str(uuid.uuid4())
    test: bool = True
    prepared: datetime = datetime.now(timezone.utc)
    sender: str = "ZZZ"
    receiver: Optional[str] = None
    source: Optional[str] = None
    dataset_action: Optional[ActionType] = None


ORGS = "OrganisationSchemes"
CLS = "Codelists"
CONCEPTS = "Concepts"

MSG_CONTENT_PKG = {
    ORGS: ItemScheme,
    CLS: Codelist,
    CONCEPTS: ConceptScheme,
}


class Message(Struct, frozen=True):
    """Message class holds the content of SDMX Message.

    Attributes:
        content (Dict[str, Any]): Content of the message. The keys are the
            content type (e.g. ``OrganisationSchemes``, ``Codelists``, etc.),
            and the values are the content objects (e.g. ``ItemScheme``,
            ``Codelist``, etc.).
    """

    content: Dict[str, Any]

    def __post_init__(self) -> None:
        """Checks if the content is valid."""
        for content_key, content_value in self.content.items():
            if content_key not in MSG_CONTENT_PKG:
                raise Invalid(
                    f"Invalid content type: {content_key}",
                    "Check the docs for the proper structure on content.",
                )

            for obj_ in content_value.values():
                if not isinstance(obj_, MSG_CONTENT_PKG[content_key]):
                    raise Invalid(
                        f"Invalid content value type: {type(obj_).__name__} "
                        f"for {content_key}",
                        "Check the docs for the proper "
                        "structure on content.",
                    )

    def __get_elements(self, type_: str) -> Dict[str, Any]:
        """Returns the elements from content."""
        if type_ in self.content:
            return self.content[type_]
        raise NotFound(
            f"No {type_} found in content",
            f"Could not find any {type_} in content.",
        )

    def __get_element_by_uid(self, type_: str, unique_id: str) -> Any:
        """Returns a specific element from content."""
        if type_ not in self.content:
            raise NotFound(
                f"No {type_} found.",
                f"Could not find any {type_} in content.",
            )

        if unique_id in self.content[type_]:
            return self.content[type_][unique_id]

        raise NotFound(
            f"No {type_} with id {unique_id} found in content",
            "Could not find the requested element.",
        )

    def get_organisation_schemes(self) -> Dict[str, ItemScheme]:
        """Returns the OrganisationScheme."""
        return self.__get_elements(ORGS)

    def get_codelists(self) -> Dict[str, Codelist]:
        """Returns the Codelist."""
        return self.__get_elements(CLS)

    def get_concept_schemes(self) -> Dict[str, ConceptScheme]:
        """Returns the Concept."""
        return self.__get_elements(CONCEPTS)

    def get_organisation_scheme_by_uid(self, unique_id: str) -> ItemScheme:
        """Returns a specific OrganisationScheme."""
        return self.__get_element_by_uid(ORGS, unique_id)

    def get_codelist_by_uid(self, unique_id: str) -> Codelist:
        """Returns a specific Codelist."""
        return self.__get_element_by_uid(CLS, unique_id)

    def get_concept_scheme_by_uid(self, unique_id: str) -> ConceptScheme:
        """Returns a specific Concept."""
        return self.__get_element_by_uid(CONCEPTS, unique_id)


class SubmissionResult(Struct, frozen=True):
    """A class to represent a Submission Result."""

    action: str
    short_urn: str
    status: str

    def __str__(self) -> str:
        """Return a string representation of the SubmissionResult."""
        return (
            f"<Submission Result - "
            f"Action: {self.action} - "
            f"Short URN: {self.short_urn} - "
            f"Status: {self.status}>"
        )



"""Build SDMX-REST queries."""

from pysdmx.api.qb.availability import (
    AvailabilityFormat,
    AvailabilityMode,
    AvailabilityQuery,
)
from pysdmx.api.qb.data import DataContext, DataFormat, DataQuery
from pysdmx.api.qb.refmeta import (
    RefMetaByMetadataflowQuery,
    RefMetaByMetadatasetQuery,
    RefMetaByStructureQuery,
    RefMetaDetail,
    RefMetaFormat,
)
from pysdmx.api.qb.schema import SchemaContext, SchemaFormat, SchemaQuery
from pysdmx.api.qb.structure import (
    StructureDetail,
    StructureFormat,
    StructureQuery,
    StructureReference,
    StructureType,
)
from pysdmx.api.qb.util import ApiVersion

__all__ = [
    "ApiVersion",
    "AvailabilityFormat",
    "AvailabilityMode",
    "AvailabilityQuery",
    "DataContext",
    "DataFormat",
    "DataQuery",
    "RefMetaByMetadataflowQuery",
    "RefMetaByMetadatasetQuery",
    "RefMetaByStructureQuery",
    "RefMetaDetail",
    "RefMetaFormat",
    "SchemaContext",
    "SchemaFormat",
    "SchemaQuery",
    "StructureDetail",
    "StructureFormat",
    "StructureQuery",
    "StructureReference",
    "StructureType",
]



"""Collection of Fusion-JSON schemas for dataflow queries."""

from typing import List, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.io.json.fusion.messages.org import FusionProviderScheme
from pysdmx.model import (
    Agency,
    Components,
    DataflowInfo,
    DataflowRef,
    DataProvider,
)


class FusionDataflowRef(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a dataflow."""

    id: str
    agency: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"

    def to_model(self) -> DataflowRef:
        """Converts a FusionDataflowRef to a standard dataflow ref."""
        return DataflowRef(
            id=self.id,
            agency=self.agency,
            name=self.names[0].value if self.names else None,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
        )


class FusionDataflow(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a dataflow."""

    id: str
    names: Sequence[FusionString]
    agency: str
    dataStructureRef: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"


class FusionDataflowMessage(Struct, frozen=True):
    """Fusion-JSON payload for /dataflow queries."""

    Dataflow: Sequence[FusionDataflow]
    DataProviderScheme: Sequence[FusionProviderScheme] = ()

    def __filter(
        self,
        df: FusionDataflow,
        agency: str,
        id_: str,
        version: str,
    ) -> bool:
        if version != "+" and version != "latest":
            return (
                df.agency == agency and df.id == id_ and df.version == version
            )
        else:
            return df.agency == agency and df.id == id_

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        prvs: List[DataProvider] = []
        for dps in self.DataProviderScheme:
            prvs.extend(dps.to_model([]))
        df = list(
            filter(
                lambda df: self.__filter(df, agency, id_, version),
                self.Dataflow,
            )
        )[0]
        return DataflowInfo(
            id=df.id,
            components=components,
            agency=Agency(df.agency),
            name=df.names[0].value,
            description=df.descriptions[0].value if df.descriptions else None,
            version=df.version,
            providers=prvs,
            dsd_ref=df.dataStructureRef,
        )



"""Collection of readers for SDMX-JSON messages."""

from pysdmx.api.fmr.reader import Deserializers
from pysdmx.io.json.sdmxjson2 import messages as msg

deserializers = Deserializers(
    agencies=msg.JsonAgencyMessage,  # type: ignore[arg-type]
    categories=msg.JsonCategorySchemeMessage,  # type: ignore[arg-type]
    codes=msg.JsonCodelistMessage,  # type: ignore[arg-type]
    concepts=msg.JsonConceptSchemeMessage,  # type: ignore[arg-type]
    dataflow=msg.JsonDataflowMessage,  # type: ignore[arg-type]
    providers=msg.JsonProviderMessage,  # type: ignore[arg-type]
    schema=msg.JsonSchemaMessage,  # type: ignore[arg-type]
    hier_assoc=msg.JsonHierarchyAssociationMessage,  # type: ignore[arg-type]
    hierarchy=msg.JsonHierarchyMessage,  # type: ignore[arg-type]
    report=msg.JsonMetadataMessage,  # type: ignore[arg-type]
    mapping=msg.JsonMappingMessage,  # type: ignore[arg-type]
    code_map=msg.JsonRepresentationMapMessage,  # type: ignore[arg-type]
)



"""Read SDMX-ML submission messages."""

from typing import Any, Dict

from pysdmx.io.xml.sdmx21.__parsing_config import (
    ACTION,
    MAINTAINABLE_OBJECT,
    REG_INTERFACE,
    STATUS,
    STATUS_MSG,
    SUBMISSION_RESULT,
    SUBMIT_STRUCTURE_RESPONSE,
    SUBMITTED_STRUCTURE,
    URN,
)
from pysdmx.model.message import SubmissionResult
from pysdmx.util import parse_urn


def handle_registry_interface(dict_info: Dict[str, Any]) -> Dict[str, Any]:
    """Handle the Registry Interface message.

    Args:
        dict_info: Dictionary with the parsed data.

    Returns:
        dict: Dictionary with the parsed data.
    """
    response = dict_info[REG_INTERFACE][SUBMIT_STRUCTURE_RESPONSE]

    result = {}
    for submission_result in response[SUBMISSION_RESULT]:
        structure = submission_result[SUBMITTED_STRUCTURE]
        action = structure[ACTION]
        urn = structure[MAINTAINABLE_OBJECT][URN]
        short_urn = str(parse_urn(urn))
        status = submission_result[STATUS_MSG][STATUS]
        result[short_urn] = SubmissionResult(action, short_urn, status)
    return result



"""Collection of SDMX-JSON schemas for organisations."""

from collections import defaultdict
from datetime import datetime
from typing import Dict, Optional, Sequence, Set

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation
from pysdmx.io.json.sdmxjson2.messages.pa import JsonProvisionAgreement
from pysdmx.model import Agency, DataflowRef, DataProvider
from pysdmx.util import parse_urn


class JsonDataProviderScheme(Struct, frozen=True):
    """SDMX-JSON payload for a data provider scheme."""

    agencyID: str
    dataProviders: Sequence[DataProvider]
    description: Optional[str] = None
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    isPartial: bool = False

    def __get_df_ref(self, ref: str) -> DataflowRef:
        a = parse_urn(ref)
        return DataflowRef(id=a.id, agency=a.agency, version=a.version)

    def to_model(
        self, pas: Sequence[JsonProvisionAgreement]
    ) -> Sequence[DataProvider]:
        """Converts a JsonDataProviderScheme to a list of Organisations."""
        if pas:
            paprs: Dict[str, Set[DataflowRef]] = defaultdict(set)
            for pa in pas:
                df = self.__get_df_ref(pa.dataflow)
                pr = pa.dataProvider[pa.dataProvider.rindex(".") + 1 :]
                paprs[pr].add(df)
            return [
                DataProvider(
                    id=p.id,
                    name=p.name,
                    description=p.description,
                    contacts=p.contacts,
                    dataflows=list(paprs[p.id]),
                )
                for p in self.dataProviders
            ]
        else:
            return self.dataProviders


class JsonDataProviderSchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of data provider schemes."""

    dataProviderSchemes: Sequence[JsonDataProviderScheme]
    provisionAgreements: Sequence[JsonProvisionAgreement] = ()

    def to_model(self) -> Sequence[DataProvider]:
        """Converts a JsonDataProviderSchemes to a list of Organisations."""
        return self.dataProviderSchemes[0].to_model(self.provisionAgreements)


class JsonProviderMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataproviderscheme queries."""

    data: JsonDataProviderSchemes

    def to_model(self) -> Sequence[DataProvider]:
        """Returns the requested list of providers."""
        return self.data.to_model()


class JsonAgencyScheme(Struct, frozen=True):
    """SDMX-JSON payload for an agency scheme."""

    agencyID: str
    agencies: Sequence[Agency]
    description: Optional[str] = None
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    isPartial: bool = False


class JsonAgencySchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of agency schemes."""

    agencySchemes: Sequence[JsonAgencyScheme]


class JsonAgencyMessage(Struct, frozen=True):
    """SDMX-JSON payload for /agencyscheme queries."""

    data: JsonAgencySchemes

    def __add_owner(self, owner: str, a: Agency) -> Agency:
        oid = f"{owner}.{a.id}" if owner != "SDMX" else a.id
        return Agency(
            id=oid, name=a.name, description=a.description, contacts=a.contacts
        )

    def to_model(self) -> Sequence[Agency]:
        """Returns the requested list of agencies."""
        return [
            self.__add_owner(self.data.agencySchemes[0].agencyID, a)
            for a in self.data.agencySchemes[0].agencies
        ]



"""Enumeration for the XML message types."""

from enum import Enum


class MessageType(Enum):
    """MessageType enumeration.

    Enumeration that withholds the Message type for writing purposes.
    """

    GenericDataSet = 1
    StructureSpecificDataSet = 2
    Structure = 3
    Error = 4
    Submission = 5



"""SDMX simplified **domain model**, expressed as **Python classes**.

This module contains data classes representing a **simplified and opinionated
subset** of the SDMX information model.
"""

from re import Pattern
from typing import Any

from pysdmx.errors import NotImplemented
from pysdmx.model.__base import (
    Agency,
    Contact,
    DataConsumer,
    DataflowRef,
    DataProvider,
    Organisation,
)
from pysdmx.model.category import Category, CategoryScheme
from pysdmx.model.code import (
    Code,
    Codelist,
    HierarchicalCode,
    Hierarchy,
    HierarchyAssociation,
)
from pysdmx.model.concept import Concept, ConceptScheme, DataType, Facets
from pysdmx.model.dataflow import (
    ArrayBoundaries,
    Component,
    Components,
    DataflowInfo,
    Role,
    Schema,
)
from pysdmx.model.dataset import SeriesInfo
from pysdmx.model.map import (
    ComponentMap,
    DatePatternMap,
    FixedValueMap,
    ImplicitComponentMap,
    MultiComponentMap,
    MultiRepresentationMap,
    MultiValueMap,
    RepresentationMap,
    StructureMap,
    ValueMap,
)
from pysdmx.model.metadata import MetadataAttribute, MetadataReport


def encoders(obj: Any) -> Any:
    """Encoders for msgspec serialization.

    The pysdmx model classes are based on msgspec `Struct` classes. These
    classes offer serialization and deserialization into various formats
    for many Python types. However, some types used in pysdmx are not
    supported. This will lead to serialization errors, unless `encoders`
    is used. See details at https://jcristharif.com/msgspec/extending.html.

    For example:

        import msgspec
        from pysdmx.model import encoders
        encoder = msgspec.json.Encoder(enc_hook=encoders)

    Args:
        obj: The object to be encoded

    Returns:
        The received object converted to supported Python types

    Raises:
        NotImplemented: In case the object type is not one of the types
            needing conversion
    """
    if isinstance(obj, Pattern):
        return f"regex:{obj.pattern}"
    elif isinstance(obj, Components):
        return list(obj)
    else:
        # Raise a NotImplemented for other types
        raise NotImplemented(
            "Unsupported", f"Objects of type {type(obj)} are not supported"
        )


__all__ = [
    "Agency",
    "ArrayBoundaries",
    "Category",
    "CategoryScheme",
    "Code",
    "Codelist",
    "Component",
    "Components",
    "ComponentMap",
    "Concept",
    "ConceptScheme",
    "Contact",
    "DataConsumer",
    "DataflowInfo",
    "DataflowRef",
    "DataType",
    "DatePatternMap",
    "DataProvider",
    "Facets",
    "HierarchicalCode",
    "Hierarchy",
    "HierarchyAssociation",
    "ImplicitComponentMap",
    "StructureMap",
    "MetadataAttribute",
    "MetadataReport",
    "MultiComponentMap",
    "MultiRepresentationMap",
    "MultiValueMap",
    "Organisation",
    "RepresentationMap",
    "Role",
    "Schema",
    "SeriesInfo",
    "ValueMap",
    "FixedValueMap",
]



"""Schemas for SDMX-JSON messages."""

from pysdmx.io.json.sdmxjson2.messages.category import (
    JsonCategorySchemeMessage,
)
from pysdmx.io.json.sdmxjson2.messages.code import (
    JsonCodelistMessage,
    JsonHierarchyAssociationMessage,
    JsonHierarchyMessage,
)
from pysdmx.io.json.sdmxjson2.messages.concept import JsonConceptSchemeMessage
from pysdmx.io.json.sdmxjson2.messages.dataflow import JsonDataflowMessage
from pysdmx.io.json.sdmxjson2.messages.map import (
    JsonMappingMessage,
    JsonRepresentationMapMessage,
)
from pysdmx.io.json.sdmxjson2.messages.org import (
    JsonAgencyMessage,
    JsonProviderMessage,
)
from pysdmx.io.json.sdmxjson2.messages.report import JsonMetadataMessage
from pysdmx.io.json.sdmxjson2.messages.schema import JsonSchemaMessage

__all__ = [
    "JsonAgencyMessage",
    "JsonCategorySchemeMessage",
    "JsonCodelistMessage",
    "JsonConceptSchemeMessage",
    "JsonDataflowMessage",
    "JsonProviderMessage",
    "JsonSchemaMessage",
    "JsonHierarchyAssociationMessage",
    "JsonHierarchyMessage",
    "JsonMetadataMessage",
    "JsonMappingMessage",
    "JsonRepresentationMapMessage",
]



"""Collection of readers for Fusion-JSON messages."""

from pysdmx.api.fmr.reader import Deserializers
from pysdmx.io.json.fusion import messages as msg

deserializers = Deserializers(
    agencies=msg.FusionAgencyMessage,  # type: ignore[arg-type]
    categories=msg.FusionCategorySchemeMessage,  # type: ignore[arg-type]
    codes=msg.FusionCodelistMessage,  # type: ignore[arg-type]
    concepts=msg.FusionConceptSchemeMessage,  # type: ignore[arg-type]
    dataflow=msg.FusionDataflowMessage,  # type: ignore[arg-type]
    providers=msg.FusionProviderMessage,  # type: ignore[arg-type]
    schema=msg.FusionSchemaMessage,  # type: ignore[arg-type]
    hier_assoc=msg.FusionHierarchyAssociationMessage,  # type: ignore[arg-type]
    hierarchy=msg.FusionHierarchyMessage,  # type: ignore[arg-type]
    report=msg.FusionMetadataMessage,  # type: ignore[arg-type]
    mapping=msg.FusionMappingMessage,  # type: ignore[arg-type]
    code_map=msg.FusionRepresentationMapMessage,  # type: ignore[arg-type]
)



"""Module for writing metadata to XML files."""

from collections import OrderedDict
from typing import Any, Dict

from pysdmx.io.xml.sdmx21.writer.__write_aux import (
    ABBR_COM,
    ABBR_MSG,
    ABBR_STR,
    add_indent,
    MSG_CONTENT_PKG,
)
from pysdmx.model.__base import (
    AnnotableArtefact,
    IdentifiableArtefact,
    Item,
    ItemScheme,
    MaintainableArtefact,
    NameableArtefact,
    VersionableArtefact,
)

ANNOTATION_WRITER = OrderedDict(
    {
        "title": "AnnotationTitle",
        "type": "AnnotationType",
        "text": "AnnotationText",
        "url": "AnnotationURL",
    }
)


def __write_annotable(annotable: AnnotableArtefact, indent: str) -> str:
    """Writes the annotations to the XML file."""
    if len(annotable.annotations) == 0:
        return ""

    child1 = indent
    child2 = add_indent(child1)
    child3 = add_indent(child2)

    outfile = f"{child1}<{ABBR_COM}:Annotations>"
    for annotation in annotable.annotations:
        if annotation.id is None:
            outfile += f"{child2}<{ABBR_COM}:Annotation>"
        else:
            outfile += (
                f"{child2}<{ABBR_COM}:Annotation " f"id={annotation.id!r}>"
            )
        outfile = outfile.replace("'", '"')

        for attr, label in ANNOTATION_WRITER.items():
            if getattr(annotation, attr, None) is not None:
                value = getattr(annotation, attr)
                value = value.replace("&", "&amp;").rstrip()
                if attr == "text":
                    head_tag = f'{ABBR_COM}:{label} xml:lang="en"'
                else:
                    head_tag = f"{ABBR_COM}:{label}"
                outfile += (
                    f"{child3}<{head_tag}>" f"{value}" f"</{ABBR_COM}:{label}>"
                )

        outfile += f"{child2}</{ABBR_COM}:Annotation>"
    outfile += f"{child1}</{ABBR_COM}:Annotations>"
    return outfile


def __write_identifiable(
    identifiable: IdentifiableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the IdentifiableArtefact to the XML file."""
    attributes = ""

    attributes += f" id={identifiable.id!r}"

    if identifiable.uri is not None:
        attributes += f" uri={identifiable.uri!r}"

    if identifiable.urn is not None:
        attributes += f" urn={identifiable.urn!r}"

    outfile = {
        "Annotations": __write_annotable(identifiable, indent),
        "Attributes": attributes,
    }

    return outfile


def __write_nameable(
    nameable: NameableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the NameableArtefact to the XML file."""
    outfile = __write_identifiable(nameable, indent)
    attrs = ["Name", "Description"]

    for attr in attrs:
        if getattr(nameable, attr.lower(), None) is not None:
            outfile[attr] = [
                (
                    f"{indent}"
                    f'<{ABBR_COM}:{attr} xml:lang="en">'
                    f"{getattr(nameable, attr.lower())}"
                    f"</{ABBR_COM}:{attr}>"
                )
            ]

    return outfile


def __write_versionable(
    versionable: VersionableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the VersionableArtefact to the XML file."""
    outfile = __write_nameable(versionable, add_indent(indent))

    outfile["Attributes"] += f" version={versionable.version!r}"

    if versionable.valid_from is not None:
        valid_from_str = versionable.valid_from.strftime("%Y-%m-%dT%H:%M:%S")
        outfile["Attributes"] += f" validFrom={valid_from_str!r}"

    if versionable.valid_to is not None:
        valid_to_str = versionable.valid_to.strftime("%Y-%m-%dT%H:%M:%S")
        outfile["Attributes"] += f" validTo={valid_to_str!r}"

    return outfile


def __write_maintainable(
    maintainable: MaintainableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the MaintainableArtefact to the XML file."""
    outfile = __write_versionable(maintainable, indent)

    outfile["Attributes"] += (
        f" isExternalReference="
        f"{str(maintainable.is_external_reference).lower()!r}"
    )

    outfile["Attributes"] += f" isFinal={str(maintainable.is_final).lower()!r}"

    if isinstance(maintainable.agency, str):
        outfile["Attributes"] += f" agencyID={maintainable.agency!r}"
    else:
        outfile["Attributes"] += f" agencyID={maintainable.agency.id!r}"

    return outfile


def __write_item(item: Item, indent: str) -> str:
    """Writes the item to the XML file."""
    head = f"{ABBR_STR}:" + type(item).__name__

    data = __write_nameable(item, add_indent(indent))
    attributes = data["Attributes"].replace("'", '"')
    outfile = f"{indent}<{head}{attributes}>"
    outfile += __export_intern_data(data, add_indent(indent))
    outfile += f"{indent}</{head}>"
    return outfile


def __write_item_scheme(item_scheme: ItemScheme, indent: str) -> str:
    """Writes the item scheme to the XML file."""
    label = f"{ABBR_STR}:{type(item_scheme).__name__}"

    data = __write_maintainable(item_scheme, indent)

    data["Attributes"] += f" isPartial={str(item_scheme.is_partial).lower()!r}"

    outfile = ""

    attributes = data.get("Attributes") or ""
    attributes = attributes.replace("'", '"')

    outfile += f"{indent}<{label}{attributes}>"

    outfile += __export_intern_data(data, indent)

    for item in item_scheme.items:
        outfile += __write_item(item, add_indent(indent))

    outfile += f"{indent}</{label}>"

    return outfile


def __write_metadata_element(
    package: Dict[str, Any], key: str, prettyprint: object
) -> str:
    """Writes the metadata element to the XML file.

    Args:
        package: The package to be written
        key: The key to be used
        prettyprint: Prettyprint or not

    Returns:
        A string with the metadata element
    """
    outfile = ""
    nl = "\n" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""

    base_indent = f"{nl}{child2}"

    if key in package:
        outfile += f"{base_indent}<{ABBR_STR}:{MSG_CONTENT_PKG[key]}>"
        for item_scheme in package[key].values():
            outfile += __write_item_scheme(
                item_scheme, add_indent(base_indent)
            )
        outfile += f"{base_indent}</{ABBR_STR}:{MSG_CONTENT_PKG[key]}>"

    return outfile


def __get_outfile(obj_: Dict[str, Any], key: str = "") -> str:
    """Generates an outfile from the object.

    Args:
        obj_: The object to be used
        key: The key to be used

    Returns:
        A string with the outfile

    """
    element = obj_.get(key) or []

    return "".join(element)


def __export_intern_data(data: Dict[str, Any], indent: str) -> str:
    """Export internal data (Annotations, Name, Description) on the XML file.

    Args:
        data: Information to be exported
        indent: Indentation used

    Returns:
        The XML string with the exported data
    """
    outfile = __get_outfile(data, "Annotations")
    outfile += __get_outfile(data, "Name")
    outfile += __get_outfile(data, "Description")

    return outfile


def generate_structures(content: Dict[str, Any], prettyprint: bool) -> str:
    """Writes the structures to the XML file.

    Args:
        content: The Message Content to be written
        prettyprint: Prettyprint or not

    Returns:
        A string with the structures
    """
    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""

    outfile = f"{nl}{child1}<{ABBR_MSG}:Structures>"

    for key in MSG_CONTENT_PKG:
        outfile += __write_metadata_element(content, key, prettyprint)

    outfile += f"{nl}{child1}</{ABBR_MSG}:Structures>"

    return outfile



"""Utility functions and constants for the parsers module."""

from typing import Any, Dict, List

# Common
ID = "id"
AGENCY_ID = "agencyID"
XMLNS = "xmlns"
VERSION = "version"

# Structure Specific
VALUE = "Value"

# Header
REF = "Ref"

# Structures
# Common
NAME = "Name"
DESC = "Description"
STR_URL = "structureURL"
STR_URL_LOW = "structure_url"
SER_URL = "serviceURL"
SER_URL_LOW = "service_url"
IS_EXTERNAL_REF = "isExternalReference"
IS_EXTERNAL_REF_LOW = "is_external_reference"
IS_FINAL = "isFinal"
IS_FINAL_LOW = "is_final"
IS_PARTIAL = "isPartial"
IS_PARTIAL_LOW = "is_partial"
# General
ANNOTATIONS = "Annotations"

# Individual
CL = "Codelist"
CON = "Concept"

# Dimension
DIM = "Dimension"

# Measure
PRIM_MEASURE = "PrimaryMeasure"

# Group Dimension
GROUP = "Group"
DIM_REF = "DimensionReference"

# Constraints
KEY_VALUE = "KeyValue"

# Annotation
ANNOTATION = "Annotation"
ANNOTATION_TITLE = "AnnotationTitle"
ANNOTATION_TYPE = "AnnotationType"
ANNOTATION_TEXT = "AnnotationText"
ANNOTATION_URL = "AnnotationURL"

TITLE = "title"
TEXT = "text"
TYPE = "type"
URL = "url"

# Facets
FACETS = "facets"
TEXT_TYPE = "textType"
TEXT_TYPE_LOW = "text_type"

# Contact
CONTACT = "Contact"
DEPARTMENT = "Department"
ROLE = "Role"
URIS = "uris"
EMAILS = "emails"
TELEPHONES = "telephones"
FAXES = "faxes"
URI = "URI"
EMAIL = "Email"
X400 = "X400"
TELEPHONE = "Telephone"
FAX = "Fax"

# Extras
AGENCY = "Agency"
PAR_ID = "maintainableParentID"
PAR_VER = "maintainableParentVersion"

# Errors
missing_rep: Dict[str, List[Any]] = {"CON": [], "CS": [], "CL": []}
dsd_id: str = ""

# Structure types
AGENCIES = "AgencyScheme"
ORGS = "OrganisationSchemes"
CLS = "Codelists"
CONCEPTS = "ConceptSchemes"
CS = "ConceptScheme"
CODE = "Code"

FacetType = {
    "minLength": "min_length",
    "maxLength": "max_length",
    "minValue": "min_value",
    "maxValue": "max_value",
    "startValue": "start_value",
    "endValue": "end_value",
    "interval": "interval",
    "timeInterval": "time_interval",
    "decimals": "decimals",
    "pattern": "pattern",
    "startTime": "start_time",
    "endTime": "end_time",
    "isSequence": "is_sequence",
}


def unique_id(agencyID: str, id_: str, version: str) -> str:
    """Create a unique ID for an object.

    Args:
        agencyID: The agency ID
        id_: The ID of the object
        version: The version of the object

    Returns:
        A string with the unique ID
    """
    return f"{agencyID}:{id_}({version})"



"""Writer auxiliary functions."""

from collections import OrderedDict
from typing import Any, Dict, Optional

from pysdmx.io.xml.enums import MessageType
from pysdmx.model.message import Header

MESSAGE_TYPE_MAPPING = {
    MessageType.GenericDataSet: "GenericData",
    MessageType.StructureSpecificDataSet: "StructureSpecificData",
    MessageType.Structure: "Structure",
}

ABBR_MSG = "mes"
ABBR_GEN = "gen"
ABBR_COM = "com"
ABBR_STR = "str"
ABBR_SPE = "ss"

ANNOTATIONS = "Annotations"
STRUCTURES = "Structures"
ORGS = "OrganisationSchemes"
AGENCIES = "AgencyScheme"
CODELISTS = "Codelists"
CONCEPTS = "Concepts"
DSDS = "DataStructures"
DATAFLOWS = "Dataflows"
CONSTRAINTS = "Constraints"

BASE_URL = "http://www.sdmx.org/resources/sdmxml/schemas/v2_1"

NAMESPACES = {
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    ABBR_MSG: f"{BASE_URL}/message",
    ABBR_GEN: f"{BASE_URL}/generic",
    ABBR_COM: f"{BASE_URL}/common",
    ABBR_STR: f"{BASE_URL}/structure",
    ABBR_SPE: f"{BASE_URL}/structureSpecific",
}

URN_DS_BASE = "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="


def __namespaces_from_type(type_: MessageType) -> str:
    """Returns the namespaces for the XML file based on type.

    Args:
        type_: MessageType to be used

    Returns:
        A string with the namespaces
    """
    return f"xmlns:{ABBR_STR}={NAMESPACES[ABBR_STR]!r} "


def create_namespaces(
    type_: MessageType, content: Dict[str, Any], prettyprint: bool = False
) -> str:
    """Creates the namespaces for the XML file.

    Args:
        type_: MessageType to be used
        content: Datasets or None
        prettyprint: Prettyprint or not

    Returns:
        A string with the namespaces
    """
    nl = "\n" if prettyprint else ""

    outfile = f'<?xml version="1.0" encoding="UTF-8"?>{nl}'

    outfile += f"<{ABBR_MSG}:{MESSAGE_TYPE_MAPPING[type_]} "
    outfile += f'xmlns:xsi={NAMESPACES["xsi"]!r} '
    outfile += f"xmlns:{ABBR_MSG}={NAMESPACES[ABBR_MSG]!r} "
    outfile += __namespaces_from_type(type_)
    outfile += (
        f"xmlns:{ABBR_COM}={NAMESPACES[ABBR_COM]!r} "
        f'xsi:schemaLocation="{NAMESPACES[ABBR_MSG]} '
        f'https://registry.sdmx.org/schemas/v2_1/SDMXMessage.xsd">'
    )

    return outfile.replace("'", '"')


MSG_CONTENT_PKG = OrderedDict(
    [
        (ORGS, "OrganisationSchemes"),
        (DATAFLOWS, "Dataflows"),
        (CODELISTS, "Codelists"),
        (CONCEPTS, "Concepts"),
        (DSDS, "DataStructures"),
        (CONSTRAINTS, "ContentConstraints"),
    ]
)


MSG_CONTENT_ITEM = {
    ORGS: "AgencyScheme",
    DATAFLOWS: "Dataflow",
    CODELISTS: "Codelist",
    CONCEPTS: "ConceptScheme",
    DSDS: "DataStructure",
    CONSTRAINTS: "ContentConstraint",
}


def get_end_message(type_: MessageType, prettyprint: bool) -> str:
    """Returns the end message for the XML file.

    Args:
        type_: MessageType to be used
        prettyprint: Prettyprint or not

    Returns:
        A string with the end message
    """
    nl = "\n" if prettyprint else ""
    return f"{nl}</{ABBR_MSG}:{MESSAGE_TYPE_MAPPING[type_]}>"


def add_indent(indent: str) -> str:
    """Adds another indent.

    Args:
        indent: The string to be indented

    Returns:
        A string with one more indentation
    """
    return indent + "\t"


def __value(element: str, value: Optional[str], prettyprint: bool) -> str:
    """Generates a value element for the XML file.

    A Value element is an XML tag with a value.

    Args:
        element: ID, Test, Prepared, Sender, Receiver, Source
        value: The value to be written
        prettyprint: Prettyprint or not

    Returns:
        A string with the value element
    """
    if not value:
        return ""
    nl = "\n" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""
    return (
        f"{nl}{child2}<{ABBR_MSG}:{element}>"
        f"{value}"
        f"</{ABBR_MSG}:{element}>"
    )


def __item(element: str, id_: Optional[str], prettyprint: bool) -> str:
    """Generates an item element for the XML file.

    An Item element is an XML tag with an id attribute.

    Args:
        element: Sender, Receiver
        id_: The ID to be written
        prettyprint: Prettyprint or not

    Returns:
        A string with the item element
    """
    if not id_:
        return ""
    nl = "\n" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""
    return f"{nl}{child2}<{ABBR_MSG}:{element} id={id_!r}/>"


def __write_header(header: Header, prettyprint: bool) -> str:
    """Writes the Header part of the message.

    Args:
        header: The Header to be written
        prettyprint: Prettyprint or not

    Returns:
        The XML string
    """
    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""
    prepared = header.prepared.strftime("%Y-%m-%dT%H:%M:%S")
    test = str(header.test).lower()
    return (
        f"{nl}{child1}<{ABBR_MSG}:Header>"
        f"{__value('ID', header.id, prettyprint)}"
        f"{__value('Test', test, prettyprint)}"
        f"{__value('Prepared', prepared, prettyprint)}"
        f"{__item('Sender', header.sender, prettyprint)}"
        f"{__item('Receiver', header.receiver, prettyprint)}"
        f"{__value('Source', header.source, prettyprint)}"
        f"{nl}{child1}</{ABBR_MSG}:Header>"
    ).replace("'", '"')



"""Collection of SDMX-JSON schemas for concepts."""

from datetime import datetime
from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.code import JsonCodelist
from pysdmx.io.json.sdmxjson2.messages.core import (
    JsonAnnotation,
    JsonRepresentation,
)
from pysdmx.model import Codelist, Concept, ConceptScheme, DataType


class IsoConceptReference(Struct, frozen=True):
    """Payload for a reference to an ISO 11179 concept."""

    conceptAgency: str
    conceptSchemeID: str
    conceptID: str


class JsonConcept(Struct, frozen=True):
    """SDMX-JSON payload for concepts."""

    id: str
    coreRepresentation: Optional[JsonRepresentation] = None
    name: Optional[str] = None
    description: Optional[str] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    parent: Optional[str] = None
    isoConceptReference: Optional[IsoConceptReference] = None

    def to_model(self, codelists: Sequence[Codelist]) -> Concept:
        """Converts a JsonConcept to a standard concept."""
        repr_ = self.coreRepresentation
        if repr_:
            if repr_.enumerationFormat:
                dt = DataType(repr_.enumerationFormat.dataType)
            elif repr_.format:
                dt = DataType(repr_.format.dataType)
            else:
                dt = DataType.STRING
            facets = repr_.to_facets()
            codes = repr_.to_enumeration(codelists, [])
            cl_ref = repr_.enumeration
        else:
            dt = DataType.STRING
            facets = None
            codes = None
            cl_ref = None
        return Concept(
            id=self.id,
            dtype=dt,
            facets=facets,
            name=self.name,
            description=self.description,
            codes=codes,
            enum_ref=cl_ref,
        )


class JsonConceptScheme(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a concept scheme."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    isPartial: bool = False
    concepts: Sequence[JsonConcept] = ()

    def to_model(self, codelists: Sequence[JsonCodelist]) -> ConceptScheme:
        """Converts a JsonConceptScheme to a standard concept scheme."""
        cls = [c.to_model() for c in codelists]
        return ConceptScheme(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[c.to_model(cls) for c in self.concepts],
        )


class JsonConceptSchemes(
    Struct,
    frozen=True,
):
    """SDMX-JSON payload for the list of concept schemes."""

    codelists: Sequence[JsonCodelist]
    conceptSchemes: Sequence[JsonConceptScheme]


class JsonConceptSchemeMessage(
    Struct,
    frozen=True,
):
    """SDMX-JSON payload for /conceptscheme queries."""

    data: JsonConceptSchemes

    def to_model(self) -> ConceptScheme:
        """Returns the requested concept scheme."""
        return self.data.conceptSchemes[0].to_model(self.data.codelists)



"""Build SDMX-REST data queries."""

from datetime import datetime
from enum import Enum
from typing import Optional, Sequence, Union

import msgspec

from pysdmx.api.dc.query import (
    MultiFilter,
    NumberFilter,
    TextFilter,
)
from pysdmx.api.qb.data import _CoreDataQuery, DataContext
from pysdmx.api.qb.structure import StructureReference
from pysdmx.api.qb.util import (
    ApiVersion,
    check_multiple_data_context,
    REST_ALL,
)
from pysdmx.errors import Invalid


class AvailabilityMode(str, Enum):
    """The mode of a data availability query."""

    AVAILABLE = "available"
    EXACT = "exact"


class AvailabilityFormat(Enum):
    """The response formats for availabillity queries."""

    SDMX_ML_2_1_STRUCTURE = "application/vnd.sdmx.structure+xml;version=2.1"
    SDMX_ML_3_0_STRUCTURE = "application/vnd.sdmx.structure+xml;version=3.0.0"
    SDMX_JSON_1_0_0 = "application/vnd.sdmx.structure+json;version=1.0.0"
    SDMX_JSON_2_0_0 = "application/vnd.sdmx.structure+json;version=2.0.0"


REFERENCES_V1 = [
    StructureReference.ALL,
    StructureReference.NONE,
    StructureReference.CODELIST,
    StructureReference.CONCEPT_SCHEME,
    StructureReference.DATAFLOW,
    StructureReference.DATA_PROVIDER_SCHEME,
    StructureReference.DATA_STRUCTURE,
]

REFERENCES_V2 = [
    StructureReference.ALL,
    StructureReference.NONE,
    StructureReference.CODELIST,
    StructureReference.CONCEPT_SCHEME,
    StructureReference.DATAFLOW,
    StructureReference.DATA_PROVIDER_SCHEME,
    StructureReference.DATA_STRUCTURE,
    StructureReference.VALUE_LIST,
]


class AvailabilityQuery(_CoreDataQuery, frozen=True, omit_defaults=True):
    """A data availability query.

    Availability queries allow to see what data are available for a
    specific context (data structure, dataflow or provision agreement).

    Attributes:
        context: The context for which data must be retrieved.
        agency_id: The agency maintaining the context to be considered.
        resource_id: The id of the context to be considered.
        version: The version of the context to be considered.
        key: The combination of dimension values identifying the slice
            of the cube for which data should be returned. Wildcarding
            is supported via the * operator.
        component_id: The id of the dimension for which to obtain availability
            information about. In the case where this information is
            not provided, data availability will be provided for all
            dimensions.
        components: The component values to be used for data filtering.
        updated_after: The last time the query was performed by the client.
        references: This attribute instructs the web service to return (or not)
            the artefacts referenced by the DataConstraint to be returned.
        mode: Return a Cube Region containing values which will be returned
            by executing the query (mode="exact") vs a Cube Region showing
            what values remain valid selections that could be added to the
            data query (mode="available").
    """

    context: DataContext = DataContext.ALL
    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_ALL
    key: Union[str, Sequence[str]] = REST_ALL
    component_id: Union[str, Sequence[str]] = REST_ALL
    components: Union[MultiFilter, None, NumberFilter, TextFilter] = None
    updated_after: Optional[datetime] = None
    references: Union[StructureReference, Sequence[StructureReference]] = (
        StructureReference.NONE
    )
    mode: AvailabilityMode = AvailabilityMode.EXACT

    def _validate_query(self, api_version: ApiVersion) -> None:
        self.validate()
        super()._validate_context(self.context, api_version)
        super()._check_multiple_contexts(
            self.agency_id,
            self.resource_id,
            self.version,
            self.key,
            api_version,
        )
        super()._check_resource_id(self.resource_id, api_version)
        super()._check_components(self.components, api_version)
        self.__validate_references(api_version)
        self.__validate_component_id(api_version)

    def __validate_reference(
        self, ref: StructureReference, api_version: ApiVersion
    ) -> None:
        if (api_version >= ApiVersion.V2_0_0 and ref not in REFERENCES_V2) or (
            api_version < ApiVersion.V2_0_0 and ref not in REFERENCES_V1
        ):
            raise Invalid(
                "Validation Error",
                f"{ref} is not allowed for SDMX-REST {api_version.value}.",
            )

    def __validate_references(self, api_version: ApiVersion) -> None:
        if isinstance(self.references, StructureReference):
            self.__validate_reference(self.references, api_version)
        else:
            refs = [r.value for r in self.references]
            check_multiple_data_context("references", refs, api_version)
            for ref in self.references:
                self.__validate_reference(ref, api_version)

    def __validate_component_id(self, api_version: ApiVersion) -> None:
        if (
            isinstance(self.component_id, (list, tuple))
            and api_version < ApiVersion.V2_0_0
        ):
            raise Invalid(
                "Validation Error",
                (
                    f"Only one component ID is allowed in SDMX-REST "
                    f"{api_version.value}."
                ),
            )

    def _get_decoder(self) -> msgspec.json.Decoder:  # type: ignore[type-arg]
        return _availability_decoder

    def __get_short_v2_qs(self) -> str:
        qs = ""
        if self.updated_after:
            qs = super()._append_qs_param(
                qs,
                self.updated_after,
                "updatedAfter",
                self.updated_after.isoformat("T", "seconds"),
            )
        if self.references != StructureReference.NONE:
            if isinstance(self.references, StructureReference):
                r = self.references.value
            else:
                refs = [ref.value for ref in self.references]
                r = ",".join(refs)
            qs = super()._append_qs_param(qs, r, "references")
        if self.mode != AvailabilityMode.EXACT:
            qs = super()._append_qs_param(qs, self.mode.value, "mode")
        return f"?{qs}" if qs else qs

    def __get_short_v1_qs(self) -> str:
        qs = ""
        if self.updated_after:
            qs = super()._append_qs_param(
                qs,
                self.updated_after,
                "updatedAfter",
                self.updated_after.isoformat("T", "seconds"),
            )
        if self.references != StructureReference.NONE:
            qs = super()._append_qs_param(
                qs,
                self.references.value,  # type: ignore[union-attr]
                "references",
            )
        if self.mode != AvailabilityMode.EXACT:
            qs = super()._append_qs_param(qs, self.mode.value, "mode")
        return f"?{qs}" if qs else qs

    def _create_full_query(self, api_version: ApiVersion) -> str:
        o = (
            "/availability"
            if api_version >= ApiVersion.V2_0_0
            else "/availableconstraint"
        )
        if api_version >= ApiVersion.V2_0_0:
            c = super()._get_v2_context_id(
                self.context,
                self.agency_id,
                self.resource_id,
                self.version,
                api_version,
            )
        else:
            c = super()._get_v1_context_id(
                self.agency_id,  # type: ignore[arg-type]
                self.resource_id,  # type: ignore[arg-type]
                self.version,  # type: ignore[arg-type]
                api_version,
            )
        o += f"{c}/{super()._to_kws(self.key, api_version)}"
        o += f"/{super()._to_kws(self.component_id, api_version)}"
        qs = ""
        if self.components:
            qs += self._create_component_filters(self.components)
        if self.updated_after:
            qs = super()._append_qs_param(
                qs,
                self.updated_after,
                "updatedAfter",
                self.updated_after.isoformat("T", "seconds"),
            )
        if isinstance(self.references, StructureReference):
            r = self.references.value
        else:
            refs = [ref.value for ref in self.references]
            r = ",".join(refs)
        qs = super()._append_qs_param(qs, r, "references")
        qs = super()._append_qs_param(qs, self.mode.value, "mode")
        return f"{o}?{qs}"

    def _create_short_query(self, api_version: ApiVersion) -> str:
        if api_version >= ApiVersion.V2_0_0:
            p = super()._get_short_v2_path(
                "availability",
                self.context,
                self.agency_id,
                self.resource_id,
                self.version,
                self.key,
                api_version,
                self.component_id,
            )
            q = self.__get_short_v2_qs()
            o = f"{p}{q}"
        else:
            p = super()._get_short_v1_path(
                "availableconstraint",
                self.agency_id,  # type: ignore[arg-type]
                self.resource_id,  # type: ignore[arg-type]
                self.version,  # type: ignore[arg-type]
                self.key,
                api_version,
            )
            q = self.__get_short_v1_qs()
            o = f"{p}{q}"
        return o


_availability_decoder = msgspec.json.Decoder(AvailabilityQuery)



"""Module that holds the necessary functions to read xml files."""

import itertools
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd

from pysdmx.errors import NotFound, NotImplemented
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.sdmx21.__parsing_config import (
    AGENCY_ID,
    ATTRIBUTES,
    DIM_OBS,
    exc_attributes,
    GROUP,
    ID,
    OBS,
    OBS_DIM,
    OBSKEY,
    OBSVALUE,
    REF,
    SERIES,
    SERIESKEY,
    STR_USAGE,
    STRID,
    STRREF,
    STRSPE,
    STRUCTURE,
    URN,
    VALUE,
    VERSION,
)
from pysdmx.io.xml.utils import add_list
from pysdmx.util import parse_urn

READING_CHUNKSIZE = 50000


def __get_element_to_list(data: Dict[str, Any], mode: Any) -> Dict[str, Any]:
    obs = {}
    data[mode][VALUE] = add_list(data[mode][VALUE])
    for k in data[mode][VALUE]:
        obs[k[ID]] = k[VALUE.lower()]
    return obs


def __process_df(
    test_list: List[Dict[str, Any]],
    df: Optional[pd.DataFrame],
    is_end: bool = False,
) -> Any:
    if not is_end and len(test_list) <= READING_CHUNKSIZE:
        return test_list, df
    if df is not None:
        df = pd.concat([df, pd.DataFrame(test_list)], ignore_index=True)
    else:
        df = pd.DataFrame(test_list)

    del test_list[:]

    return test_list, df


def __reading_generic_series(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Generic Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for series in dataset[SERIES]:
        keys = {}
        # Series Keys
        series[SERIESKEY][VALUE] = add_list(series[SERIESKEY][VALUE])
        for v in series[SERIESKEY][VALUE]:
            keys[v[ID]] = v[VALUE.lower()]
        if ATTRIBUTES in series:
            series[ATTRIBUTES][VALUE] = add_list(series[ATTRIBUTES][VALUE])
            for v in series[ATTRIBUTES][VALUE]:
                keys[v[ID]] = v[VALUE.lower()]
        if OBS in series:
            series[OBS] = add_list(series[OBS])

            for data in series[OBS]:
                obs = {
                    OBS_DIM: data[OBS_DIM][VALUE.lower()],
                    OBSVALUE.upper(): data[OBSVALUE][VALUE.lower()],
                }
                if ATTRIBUTES in data:
                    obs = {
                        **obs,
                        **__get_element_to_list(data, mode=ATTRIBUTES),
                    }
                test_list.append({**keys, **obs})
        else:
            test_list.append(keys)
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def __reading_generic_all(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Generic All Dimensions
    test_list = []
    df = None
    dataset[OBS] = add_list(dataset[OBS])
    for data in dataset[OBS]:
        obs: Dict[str, Any] = {}
        obs = {
            **obs,
            **__get_element_to_list(data, mode=OBSKEY),
            OBSVALUE.upper(): data[OBSVALUE][VALUE.lower()],
        }
        if ATTRIBUTES in data:
            obs = {**obs, **__get_element_to_list(data, mode=ATTRIBUTES)}
        test_list.append({**obs})
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def __reading_str_series(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Structure Specific Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for data in dataset[SERIES]:
        keys = dict(itertools.islice(data.items(), len(data)))
        if OBS in data:
            del keys[OBS]
            data[OBS] = add_list(data[OBS])
            for j in data[OBS]:
                test_list.append({**keys, **j})
        else:
            test_list.append(keys)
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def __reading_group_data(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Structure Specific Group Data
    test_list = []
    df = None
    dataset[GROUP] = add_list(dataset[GROUP])
    for data in dataset[GROUP]:
        test_list.append(dict(data.items()))
        test_list, df = __process_df(test_list, df)
    test_list, df = __process_df(test_list, df, is_end=True)

    cols_to_delete = [x for x in df.columns if ":type" in x]
    for x in cols_to_delete:
        del df[x]

    df = df.drop_duplicates(keep="first").reset_index(drop=True)

    return df


def __get_at_att_str(dataset: Dict[str, Any]) -> Dict[str, Any]:
    """Gets the elements of the dataset if it is Structure Specific Data."""
    return {k: dataset[k] for k in dataset if k not in exc_attributes}


def __get_at_att_gen(dataset: Dict[str, Any]) -> Dict[str, Any]:
    """Gets all the elements if it is Generic data."""
    attached_attributes: Dict[str, Any] = {}
    if ATTRIBUTES not in dataset:
        return attached_attributes
    dataset[ATTRIBUTES][VALUE] = add_list(dataset[ATTRIBUTES][VALUE])
    for k in dataset[ATTRIBUTES][VALUE]:
        attached_attributes[k[ID]] = k[VALUE.lower()]
    return attached_attributes


def __get_ids_from_structure(element: Dict[str, Any]) -> Any:
    """Gets the agency_id, id and version of the structure.

    Args:
        element: The data hold in the structure.

    Returns:
        If the element is REF, agency_id, id and version may be returned.
        If the element is URN, agency_id, id and version would be taken from
        split function.
    """
    if REF in element:
        agency_id = element[REF][AGENCY_ID]
        id_ = element[REF][ID]
        version = element[REF][VERSION]
        return agency_id, id_, version
    else:
        urn = parse_urn(element[URN])
        return urn.agency, urn.id, urn.version


def __get_elements_from_structure(structure: Dict[str, Any]) -> Any:
    """Gets elements according to the xml type of file.

    Args:
        structure: It can appear in two ways:
            If structure is 'STRUCTURE', it will get
            the ids related to STRUCTURE.
            If structure is 'STR_USAGE', it will get
            the ids related to STR_USAGE.

    Returns:
        The ids contained in the structure will be returned.

    Raises:
        NotImplemented: For Provision Agreement, as it is not implemented.
    """
    if STRUCTURE in structure:
        structure_type = "DataStructure"
        tuple_ids = __get_ids_from_structure(structure[STRUCTURE])

    elif STR_USAGE in structure:
        structure_type = "DataFlow"
        tuple_ids = __get_ids_from_structure(structure[STR_USAGE])
    else:
        raise NotImplemented(
            "Unsupported", "ProvisionAgrement not implemented"
        )
    return tuple_ids + (structure_type,)


def __extract_structure(structure: Any) -> Any:
    """Extracts elements contained in the structure."""
    structure = add_list(structure)
    str_info = {}
    for str_item in structure:
        (agency_id, id_, version, structure_type) = (
            __get_elements_from_structure(str_item)
        )

        str_id = f"{agency_id}:{id_}({version})"

        str_info[str_item[STRID]] = {
            DIM_OBS: str_item[DIM_OBS],
            "unique_id": str_id,
            "structure_type": structure_type,
        }

    return str_info


def __parse_structure_specific_data(
    dataset: Dict[str, Any], structure_info: Dict[str, Any]
) -> PandasDataset:
    attached_attributes = __get_at_att_str(dataset)

    # Parsing data
    if SERIES in dataset:
        # Structure Specific Series
        df = __reading_str_series(dataset)
        if GROUP in dataset:
            df_group = __reading_group_data(dataset)
            common_columns = list(
                set(df.columns).intersection(set(df_group.columns))
            )
            df = pd.merge(df, df_group, on=common_columns, how="left")
    else:
        dataset[OBS] = add_list(dataset[OBS])
        # Structure Specific All dimensions
        df = pd.DataFrame(dataset[OBS]).replace(np.nan, "")

    urn = (
        "urn:sdmx:org.sdmx.infomodel.datastructure."
        f"{structure_info['structure_type']}={structure_info['unique_id']}"
    )

    return PandasDataset(
        structure=urn, attributes=attached_attributes, data=df
    )


def __parse_generic_data(
    dataset: Dict[str, Any], structure_info: Dict[str, Any]
) -> PandasDataset:
    attached_attributes = __get_at_att_gen(dataset)

    # Parsing data
    if SERIES in dataset:
        # Generic Series
        df = __reading_generic_series(dataset)
    else:
        # Generic All Dimensions
        df = __reading_generic_all(dataset)

    urn = (
        "urn:sdmx:org.sdmx.infomodel.datastructure."
        f"{structure_info['structure_type']}={structure_info['unique_id']}"
    )

    return PandasDataset(
        structure=urn, attributes=attached_attributes, data=df
    )


def create_dataset(
    dataset: Any, str_info: Dict[str, Any], global_mode: Any
) -> PandasDataset:
    """Creates the dataset from the xml file.

    Takes the information contained in the xml files
    to fulfill the dataset and return a pandas dataframe.

    Args:
        dataset: The dataset created.
        str_info: Dict that contains info
            such as agency_id, id and its version.
        global_mode: Identifies if the xml file has
            Generic data or a StructureSpecificData.

    Returns:
        A pandas dataframe with the created dataset will be returned.

    Raises:
        NotFound: If the structure reference cannot be found.
    """
    if dataset[STRREF] not in str_info:
        raise NotFound(
            "Unknown structure",
            f"Cannot find the structure reference "
            f"of this dataset:{dataset[STRREF]}",
        )
    structure_info = str_info[dataset[STRREF]]
    if STRSPE == global_mode:
        return __parse_structure_specific_data(dataset, structure_info)
    else:
        return __parse_generic_data(dataset, structure_info)



"""Processes the input that comes into read_sdmx function."""

from io import BytesIO, TextIOWrapper
from json import JSONDecodeError, loads
from os import PathLike
from pathlib import Path
from typing import Tuple, Union

from pysdmx.errors import Invalid


def __remove_bom(input_string: str) -> str:
    return input_string.replace("\ufeff", "")


def __check_xml(infile: str) -> bool:
    if infile[:5] == "<?xml":
        return True

    return False


def process_string_to_read(
    infile: Union[str, Path, BytesIO]
) -> Tuple[str, str]:
    """Processes the input that comes into read_sdmx function.

    Args:
        infile: Path to file, URL, or string.

    Returns:
        tuple: Tuple containing the parsed input and the format of the input.

    Raises:
        Invalid: If the input cannot be parsed as SDMX.
    """
    # Read file as string
    if isinstance(infile, (Path, PathLike)):
        with open(infile, "r", encoding="utf-8-sig", errors="replace") as f:
            out_str = f.read()

    # Read from BytesIO
    elif isinstance(infile, BytesIO):
        text_wrap = TextIOWrapper(infile, encoding="utf-8", errors="replace")
        out_str = text_wrap.read()

    elif isinstance(infile, str):
        out_str = infile
    else:
        raise Invalid(
            "Validation Error", f"Cannot parse input of type {type(infile)}."
        )

    out_str = __remove_bom(out_str)

    # Check if string is a valid JSON
    try:
        loads(out_str)
        return out_str, "json"
    except JSONDecodeError:
        pass

    # Check if string is a valid XML
    if __check_xml(out_str):
        return out_str, "xml"

    raise Invalid(
        "Validation Error", f"Cannot parse input as SDMX. Found {infile}"
    )



"""Classes and parsers for data queries."""

from pysdmx.api.dc.query._model import (
    BooleanFilter,
    DateTimeFilter,
    LogicalOperator,
    MultiFilter,
    NotFilter,
    NullFilter,
    NumberFilter,
    Operator,
    SortBy,
    TextFilter,
)


__all__ = [
    "BooleanFilter",
    "DateTimeFilter",
    "LogicalOperator",
    "MultiFilter",
    "NotFilter",
    "NullFilter",
    "NumberFilter",
    "Operator",
    "SortBy",
    "TextFilter",
]



"""Collection of SDMX-JSON schemas for provision agreements."""

from datetime import datetime
from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation


class JsonProvisionAgreement(
    Struct, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for a provision agreement."""

    id: str
    name: str
    agency: str
    dataflow: str
    dataProvider: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None



"""Build SDMX-REST structure queries."""

from abc import abstractmethod
from enum import Enum
from typing import Sequence, Union

import msgspec
from msgspec.json import Decoder

from pysdmx.api.qb.structure import _API_RESOURCES, StructureType
from pysdmx.api.qb.util import (
    ApiVersion,
    REST_ALL,
    REST_LATEST,
)
from pysdmx.errors import Invalid


class RefMetaDetail(Enum):
    """The desired amount of information to be returned."""

    FULL = "full"
    ALL_STUBS = "allstubs"


class RefMetaFormat(Enum):
    """The response formats."""

    SDMX_ML_3_0_STRUCTURE = "application/vnd.sdmx.metadata+xml;version=3.0.0"
    SDMX_JSON_2_0_0 = "application/vnd.sdmx.metadata+json;version=2.0.0"
    SDMX_CSV_2_0_0 = "application/vnd.sdmx.metadata+csv;version=2.0.0"


class _RefMetaCoreQuery(
    msgspec.Struct,
    frozen=True,
    omit_defaults=True,
):
    def get_url(self, version: ApiVersion, omit_defaults: bool = False) -> str:
        """The URL for the query in the selected SDMX-REST API version."""
        self._validate_query(version)
        if omit_defaults:
            return self._create_short_query()
        else:
            return self._create_full_query()

    def validate(self) -> None:
        """Validate the query."""
        try:
            self._get_decoder().decode(_encoder.encode(self))
        except msgspec.DecodeError as err:
            raise Invalid(
                "Invalid Reference Metadata Query", str(err)
            ) from err

    def _check_version(self, version: ApiVersion) -> None:
        if version < ApiVersion.V2_0_0:
            raise Invalid(
                "Invalid Request",
                (
                    "Queries for reference metadata are not supported"
                    f"in SDMX-REST {version.value}."
                ),
            )

    def _join_mult(self, vals: Union[str, Sequence[str]]) -> str:
        return vals if isinstance(vals, str) else ",".join(vals)

    @abstractmethod
    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        """Returns the decoder to be used for validation."""

    @abstractmethod
    def _validate_query(self, version: ApiVersion) -> None:
        """Any additional validation steps to be performed by subclasses."""

    @abstractmethod
    def _create_full_query(self) -> str:
        """Creates a URL, with default values."""

    @abstractmethod
    def _create_short_query(self) -> str:
        """Creates a URL, omitting default values when possible."""


class RefMetaByMetadatasetQuery(
    _RefMetaCoreQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for reference metadata with metadataset identification details.

    Attributes:
        provider_id: The id(s) of the data provider.
        metadataset_id: The id(s) of the metadataset(s) to be returned.
        version: The version(s) of the metadataset(s) to be returned.
        detail: The desired amount of information to be returned.
    """

    provider_id: Union[str, Sequence[str]] = REST_ALL
    metadataset_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    detail: RefMetaDetail = RefMetaDetail.FULL

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_mds_decoder

    def _create_full_query(self) -> str:
        p = super()._join_mult(self.provider_id)
        i = super()._join_mult(self.metadataset_id)
        v = super()._join_mult(self.version)
        return f"/metadata/metadataset/{p}/{i}/{v}?detail={self.detail.value}"

    def _create_short_query(self) -> str:
        v = f"/{self.version}" if self.version != REST_LATEST else ""
        i = (
            f"/{self.metadataset_id}{v}"
            if v or self.metadataset_id != REST_ALL
            else ""
        )
        p = (
            f"/{self.provider_id}{i}"
            if i or self.provider_id != REST_ALL
            else ""
        )
        d = f"?{self.detail}" if self.detail != RefMetaDetail.FULL else ""
        return f"/metadata/metadataset{p}{d}"


class RefMetaByStructureQuery(
    _RefMetaCoreQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for reference metadata reported against one or more structures.

    Attributes:
        artefact_type: The type of structural metadata to which the
            reference metadata to be returned are attached.
        agency_id: The agency (or agencies) maintaining the artefact(s)
            to which the reference metadata to be returned are attached.
        resource_id: The id(s) of the artefact(s) to which the reference
            metadata to be returned are attached.
        version: The version(s) of the artefact(s) to which the reference
            metadata to be returned are attached.
        detail: The desired amount of information to be returned.
    """

    artefact_type: StructureType = StructureType.ALL
    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    detail: RefMetaDetail = RefMetaDetail.FULL

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)
        self.__check_artefact_type(self.artefact_type, version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_struct_decoder

    def __check_artefact_type(
        self, atyp: StructureType, version: ApiVersion
    ) -> None:
        if atyp not in _API_RESOURCES[version.name.replace("_", ".")]:
            raise Invalid(
                "Validation Error",
                f"{atyp} is not valid for SDMX-REST {version.name}.",
            )

    def _create_full_query(self) -> str:
        a = super()._join_mult(self.agency_id)
        r = super()._join_mult(self.resource_id)
        v = super()._join_mult(self.version)
        return (
            f"/metadata/structure/{self.artefact_type.value}/{a}/{r}/{v}"
            f"?detail={self.detail.value}"
        )

    def _create_short_query(self) -> str:
        v = f"/{self.version}" if self.version != REST_LATEST else ""
        r = (
            f"/{self.resource_id}{v}"
            if v or self.resource_id != REST_ALL
            else ""
        )
        a = f"/{self.agency_id}{r}" if r or self.agency_id != REST_ALL else ""
        t = (
            f"/{self.artefact_type.value}{a}"
            if a or self.artefact_type != StructureType.ALL
            else ""
        )
        d = f"?{self.detail}" if self.detail != RefMetaDetail.FULL else ""
        return f"/metadata/structure{t}{d}"


class RefMetaByMetadataflowQuery(
    _RefMetaCoreQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for reference metadata reported for metadataflows.

    Attributes:
        agency_id: The agency (or agencies) maintaining the metadataflow(s)
            of which reference metadata need to be returned.
        resource_id: The id(s) of the metadataflow(s) of which reference
            metadata need to be returned.
        version: The version(s) of the metadataflows(s) of which reference
            metadata need to be returned.
        provider_id: The id(s) of the providers that provided the reference
            metadata to be returned.
        detail: The desired amount of information to be returned.
    """

    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    provider_id: Union[str, Sequence[str]] = REST_ALL
    detail: RefMetaDetail = RefMetaDetail.FULL

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_flow_decoder

    def _create_full_query(self) -> str:
        a = super()._join_mult(self.agency_id)
        r = super()._join_mult(self.resource_id)
        v = super()._join_mult(self.version)
        p = super()._join_mult(self.provider_id)
        return (
            f"/metadata/metadataflow/{a}/{r}/{v}/{p}"
            f"?detail={self.detail.value}"
        )

    def _create_short_query(self) -> str:
        p = f"/{self.provider_id}" if self.provider_id != REST_ALL else ""
        v = f"/{self.version}{p}" if p or self.version != REST_LATEST else ""
        r = (
            f"/{self.resource_id}{v}"
            if v or self.resource_id != REST_ALL
            else ""
        )
        a = f"/{self.agency_id}{r}" if r or self.agency_id != REST_ALL else ""
        d = f"?{self.detail}" if self.detail != RefMetaDetail.FULL else ""
        return f"/metadata/metadataflow{a}{d}"


_by_mds_decoder = msgspec.json.Decoder(RefMetaByMetadatasetQuery)
_by_struct_decoder = msgspec.json.Decoder(RefMetaByStructureQuery)
_by_flow_decoder = msgspec.json.Decoder(RefMetaByMetadataflowQuery)
_encoder = msgspec.json.Encoder()


__all__ = [
    "RefMetaDetail",
    "RefMetaFormat",
    "RefMetaByStructureQuery",
]



"""Build SDMX-REST schema queries."""

from enum import Enum
from typing import Optional

import msgspec

from pysdmx.api.qb.structure import _V2_0_ADDED, StructureType
from pysdmx.api.qb.util import ApiVersion, REST_ALL, REST_LATEST
from pysdmx.errors import Invalid


class SchemaContext(Enum):
    """The context for which a schema must be generated."""

    DATA_STRUCTURE = "datastructure"
    METADATA_STRUCTURE = "metadatastructure"
    DATAFLOW = "dataflow"
    METADATA_FLOW = "metadataflow"
    PROVISION_AGREEMENT = "provisionagreement"
    METADATA_PROVISION_AGREEMENT = "metadataprovisionagreement"


class SchemaFormat(Enum):
    """The response formats."""

    SDMX_JSON_1_0_0_SCHEMA = "application/vnd.sdmx.schema+json;version=1.0.0"
    SDMX_JSON_2_0_0_SCHEMA = "application/vnd.sdmx.schema+json;version=2.0.0"
    SDMX_JSON_1_0_0_STRUCTURE = (
        "application/vnd.sdmx.structure+json;version=1.0.0"
    )
    SDMX_JSON_2_0_0_STRUCTURE = (
        "application/vnd.sdmx.structure+json;version=2.0.0"
    )
    SDMX_ML_2_1_SCHEMA = "application/vnd.sdmx.schema+xml;version=2.1"
    SDMX_ML_3_0_SCHEMA = "application/vnd.sdmx.schema+xml;version=3.0.0"
    SDMX_ML_2_1_STRUCTURE = "application/vnd.sdmx.structure+xml;version=2.1"
    SDMX_ML_3_0_STRUCTURE = "application/vnd.sdmx.structure+xml;version=3.0.0"


class SchemaQuery(msgspec.Struct, frozen=True, omit_defaults=True):
    """A schema query.

    Schema queries allow retrieving the definition of data validity for a
    certain context. The service must take into account the constraints
    that apply within that context (e.g. dataflow).

    Attributes:
        context: The context for which a schema must be generated. This
            determines the constraints that will be taken into
            consideration.
        agency_id: The agency maintaining the context to be considered.
        resource_id: The id of the context to be considered.
        version: The version of the context to be considered.
        obs_dimension: The ID of the dimension at the observation level.
    """

    context: SchemaContext
    agency_id: str
    resource_id: str
    version: str = REST_LATEST
    obs_dimension: Optional[str] = None
    explicit: bool = False

    def validate(self) -> None:
        """Validate the query."""
        try:
            decoder.decode(encoder.encode(self))
        except msgspec.DecodeError as err:
            raise Invalid("Invalid Schema Query", str(err)) from err

    def get_url(self, version: ApiVersion, omit_defaults: bool = False) -> str:
        """The URL for the query in the selected SDMX-REST API version."""
        self.__validate_query(version)
        if omit_defaults:
            return self.__create_short_query(version)
        else:
            return self.__create_full_query(version)

    def __to_kw(self, val: str, ver: ApiVersion) -> str:
        if val == "~" and ver < ApiVersion.V2_0_0:
            val = "latest"
        return val

    def __check_context(self, version: ApiVersion) -> None:
        ct = StructureType(self.context.value)
        if version < ApiVersion.V2_0_0 and ct in _V2_0_ADDED:
            raise Invalid(
                "Validation Error",
                f"{self.context} not allowed in {version.value}.",
            )

    def __check_version(self) -> None:
        if self.version == REST_ALL:
            raise Invalid(
                "Validation Error",
                "Retrieving schemas for all versions is not allowed.",
            )

    def __check_explicit(self, version: ApiVersion) -> None:
        if self.explicit and version >= ApiVersion.V2_0_0:
            raise Invalid(
                "Validation Error",
                f"Explicit parameter is not supported in {version.value}.",
            )

    def __validate_query(self, version: ApiVersion) -> None:
        self.validate()
        self.__check_context(version)
        self.__check_version()
        self.__check_explicit(version)

    def __create_full_query(self, ver: ApiVersion) -> str:
        u = (
            f"/schema/{self.context.value}/"
            f"{self.agency_id}/{self.resource_id}"
        )
        u += f"/{self.__to_kw(self.version, ver)}"
        if self.obs_dimension or ver < ApiVersion.V2_0_0:
            u += "?"
        if self.obs_dimension:
            u += f"dimensionAtObservation={self.obs_dimension}"
        if ver < ApiVersion.V2_0_0:
            if self.obs_dimension:
                u += "&"
            u += f"explicit={str(self.explicit).lower()}"
        return u

    def __create_short_query(self, ver: ApiVersion) -> str:
        u = (
            f"/schema/{self.context.value}/"
            f"{self.agency_id}/{self.resource_id}"
        )
        if self.version != REST_LATEST:
            u += f"/{self.__to_kw(self.version, ver)}"
        if self.obs_dimension or self.explicit:
            u += "?"
        if self.obs_dimension:
            u += f"dimensionAtObservation={self.obs_dimension}"
        if self.explicit:
            if self.obs_dimension:
                u += "&"
            u += f"explicit={str(self.explicit).lower()}"
        return u


decoder = msgspec.json.Decoder(SchemaQuery)
encoder = msgspec.json.Encoder()


__all__ = ["SchemaContext", "SchemaFormat", "SchemaQuery"]



"""Model for SDMX Reference Metadata.

Reference metadata are quite generic (and therefore powerful) in SDMX.
Though they are typically used to provide additional information about
statistical data, such as information about the general methodological
and quality aspects of the statistical production process, they can also
be used to drive process steps such as validation or mapping, for
example by providing configuration details in a metadata report.
"""

from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional, Sequence

from msgspec import Struct


class MetadataAttribute(Struct, frozen=True, omit_defaults=True):
    """An entry in a metadata report.

    An attribute is iterable, as it may contain other attributes.

    Attributes:
        id: The identifier of the attribute (e.g. "License").
        value: The attribute value (e.g. "BSD-2-Clause").
        attributes: The list of "children" attributes (i.e. attributes
            can be nested).
    """

    id: str
    value: Optional[Any] = None
    attributes: Sequence["MetadataAttribute"] = ()

    def __iter__(self) -> Iterator["MetadataAttribute"]:
        """Return an iterator over the list of attributes."""
        yield from self.attributes

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        return f"{self.id}: {self.value}"


class MetadataReport(Struct, frozen=True, omit_defaults=True):
    """An organized collection of metadata.

    A metadata report is iterable and it is also possible to directly
    retrieve an attribute using its ID.

    Attributes:
        id: The identifier of the report (e.g. DTI_MACRO).
        name: The name of the report (e.g. "Configuration metadata for
            the MACRO dataflow").
        metadataflow: The URN of the dataflow for which the report
            belongs.
        targets: The URN(s) of SDMX artefact(s) to which the report relates.
        attributes: The list of metadata attributes included in the report.
            Attributes may contain other attributes.
    """

    id: str
    name: str
    metadataflow: str
    targets: Sequence[str]
    attributes: Sequence[MetadataAttribute]

    def __iter__(self) -> Iterator[MetadataAttribute]:
        """Return an iterator over the list of report attributes."""
        yield from self.attributes

    def __len__(self) -> int:
        """Return the number of attributes in the report."""
        return self.__get_count(self.attributes)

    def __getitem__(self, id_: str) -> Optional[MetadataAttribute]:
        """Return the attribute identified by the given ID."""
        return self.__extract_attr(self.attributes, id_)

    def __get_count(self, attributes: Sequence[MetadataAttribute]) -> int:
        """Return the number of attributes at all levels."""
        count = len(attributes)
        for attr in attributes:
            if attr.attributes:
                count += self.__get_count(attr.attributes)
        return count

    def __extract_attr(
        self, attributes: Sequence[MetadataAttribute], id_: str
    ) -> Optional[MetadataAttribute]:
        if "." in id_:
            ids = id_.split(".")
            out = list(filter(lambda attr: attr.id == ids[0], attributes))
            if out:
                pkey = ".".join(ids[1:])
                return self.__extract_attr(out[0].attributes, pkey)
        else:
            out = list(filter(lambda attr: attr.id == id_, attributes))
            if out:
                return out[0]
        return None


def merge_attributes(
    attrs: Sequence[MetadataAttribute],
) -> Sequence[MetadataAttribute]:
    """Groups together the values of attributes with the same ID.

    The function assumes that an attribute will either have a
    value or will act as a container for other attributes. In
    case the attribute contains other attributes AND has a value,
    this function will NOT work as expected.

    Args:
        attrs: The list of attributes to be merged

    Returns:
        The list of (possibly merged) attributes
    """
    by_id: Dict[str, List[Any]] = defaultdict(list)
    sub_id = []

    for attr in attrs:
        if attr.attributes:
            sub_id.append(
                MetadataAttribute(
                    attr.id,
                    attr.value,
                    merge_attributes(attr.attributes),
                )
            )
        else:
            by_id[attr.id].append(attr.value)

    out = []
    out.extend(sub_id)
    for k, v in by_id.items():
        val = v if len(v) > 1 else v[0]
        out.append(MetadataAttribute(k, val))
    return out



"""Objects related to data queries."""

from datetime import datetime
from enum import Enum
from typing import Literal, Optional, Sequence, Union

from msgspec import Struct


class SortBy(Struct, frozen=True):
    """The desired sort order.

    Attributes:
        component: The ID of the component to be used for sorting.
        order: The sorting order (ascending or descending).
    """

    component: str
    order: Literal["asc", "desc"] = "asc"


class Operator(Enum):
    """The list of operators."""

    EQUALS = "="
    NOT_EQUALS = "<>"
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN_OR_EQUAL = ">="
    LIKE = "LIKE"
    NOT_LIKE = "NOT LIKE"
    IN = "IN"
    NOT_IN = "NOT IN"
    BETWEEN = "BETWEEN"
    NOT_BETWEEN = "NOT BETWEEN"
    NULL = "IS NULL"
    NOT_NULL = "IS NOT NULL"


class LogicalOperator(Enum):
    """The list of logical operators."""

    AND = "AND"
    OR = "OR"


class TextFilter(Struct, frozen=True, tag=True):
    """A filter on a text field (char, varchar, etc.)."""

    field: str
    operator: Operator
    value: Union[Sequence[str], str]


class NumberFilter(Struct, frozen=True, tag=True):
    """A filter on a number field (int, long, float, double, etc)."""

    field: str
    operator: Operator
    value: Union[Sequence[Union[int, float]], int, float]


class DateTimeFilter(Struct, frozen=True, tag=True):
    """A filter on a date or datetime field."""

    field: str
    operator: Operator
    value: Union[Sequence[datetime], datetime]


class BooleanFilter(Struct, frozen=True, tag=True):
    """A filter on a boolean field."""

    field: str
    operator: Operator
    value: Optional[bool]


class NullFilter(Struct, frozen=True, tag=True):
    """A filter on null values."""

    field: str
    operator: Operator
    type_hint: Optional[Literal["text", "number", "datetime", "boolean"]] = (
        None
    )


Filter = Union[
    BooleanFilter,
    DateTimeFilter,
    "MultiFilter",
    "NotFilter",
    NullFilter,
    NumberFilter,
    TextFilter,
]


class MultiFilter(Struct, frozen=True, tag=True):
    """A combination of data filters.

    This is useful to combine several conditions, for example
    to indicate that the value for a component must start with an A
    and end with a 10 (e.g. AZB56710).

    Attributes:
        filters: The list of conditions to be applied.
        operator: Whether the conditions are cumulative (AND) or
            exclusive (OR).
    """

    filters: Sequence[Filter]
    operator: LogicalOperator = LogicalOperator.AND


class NotFilter(Struct, frozen=True, tag=True):
    """A filter matching anything but the supplied filter."""

    filter: Filter


__all__ = [
    "BooleanFilter",
    "DateTimeFilter",
    "Filter",
    "MultiFilter",
    "NotFilter",
    "NullFilter",
    "NumberFilter",
    "Operator",
    "SortBy",
    "TextFilter",
]



"""Collection of SDMX-JSON schemas for reference metadata queries."""

from typing import Sequence

from msgspec import Struct

from pysdmx.model.metadata import merge_attributes, MetadataReport


class JsonMetadataSets(Struct, frozen=True):
    """SDMX-JSON payload for the list of metadata sets."""

    metadataSets: Sequence[MetadataReport]


class JsonMetadataMessage(Struct, frozen=True):
    """SDMX-JSON payload for /metadata queries."""

    data: JsonMetadataSets

    def __create_report(self, r: MetadataReport) -> MetadataReport:
        attrs = merge_attributes(r.attributes)
        return MetadataReport(r.id, r.name, r.metadataflow, r.targets, attrs)

    def to_model(self, fetch_all: bool = False) -> Sequence[MetadataReport]:
        """Returns the requested metadata report(s)."""
        if fetch_all:
            return [self.__create_report(r) for r in self.data.metadataSets]
        else:
            return [self.__create_report(self.data.metadataSets[0])]



"""Collection of SDMX-JSON schemas for codes and codelists."""

from datetime import datetime, timezone as tz
from typing import Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation, JsonLink
from pysdmx.model import (
    Code,
    Codelist,
    HierarchicalCode,
    Hierarchy,
    HierarchyAssociation,
)
from pysdmx.util import find_by_urn, parse_item_urn


class JsonCode(Struct, frozen=True):
    """SDMX-JSON payload for codes."""

    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    parent: Optional[str] = None

    def __handle_date(self, datestr: str) -> datetime:
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S%z")

    def __get_val(
        self, a: JsonAnnotation
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        vals = a.title.split("/")  # type: ignore[union-attr]
        if a.title.startswith("/"):  # type: ignore[union-attr]
            return (None, self.__handle_date(vals[1]))
        else:
            valid_from = self.__handle_date(vals[0])
            valid_to = self.__handle_date(vals[1]) if vals[1] else None
            return (valid_from, valid_to)

    def to_model(self) -> Code:
        """Converts a JsonCode to a standard code."""
        if self.annotations:
            vp = [
                a for a in self.annotations if a.type == "FR_VALIDITY_PERIOD"
            ]
        else:
            vp = None
        vf, vt = self.__get_val(vp[0]) if vp else (None, None)
        return Code(
            id=self.id,
            name=self.name,
            description=self.description,
            valid_from=vf,
            valid_to=vt,
        )


class JsonCodelist(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a codelist."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    isPartial: bool = False
    codes: Sequence[JsonCode] = ()

    def to_model(self) -> Codelist:
        """Converts a JsonCodelist to a standard codelist."""
        return Codelist(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[i.to_model() for i in self.codes],
        )


class JsonValuelist(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a valuelist."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    isPartial: bool = False
    valueItems: Sequence[JsonCode] = ()

    def to_model(self) -> Codelist:
        """Converts a JsonValuelist to a standard codelist."""
        return Codelist(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[i.to_model() for i in self.valueItems],
            sdmx_type="valuelist",
        )


class JsonCodelists(Struct, frozen=True):
    """SDMX-JSON payload for lists of codes."""

    codelists: Sequence[JsonCodelist] = ()
    valuelists: Sequence[JsonValuelist] = ()


class JsonCodelistMessage(Struct, frozen=True):
    """SDMX-JSON payload for /codelist queries."""

    data: JsonCodelists

    def to_model(self) -> Codelist:
        """Returns the requested codelist."""
        if self.data.codelists:
            return self.data.codelists[0].to_model()
        else:
            return self.data.valuelists[0].to_model()


class JsonHierarchicalCode(Struct, frozen=True):
    """Fusion-JSON payload for hierarchical codes."""

    id: str
    code: str
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    hierarchicalCodes: Sequence["JsonHierarchicalCode"] = ()

    def __find_code(self, codelists: Sequence[Codelist], urn: str) -> Code:
        r = parse_item_urn(urn)
        f = [
            c
            for c in codelists
            if (
                c.agency == r.agency
                and c.id == r.id
                and c.version == r.version
            )
        ]
        return [c for c in f[0].codes if c.id == r.item_id][0]

    def to_model(self, codelists: Sequence[Codelist]) -> HierarchicalCode:
        """Converts a JsonHierarchicalCode to a hierachical code."""
        code = self.__find_code(codelists, self.code)
        codes = [c.to_model(codelists) for c in self.hierarchicalCodes]
        vf = self.validFrom.replace(tzinfo=tz.utc) if self.validFrom else None
        vt = self.validTo.replace(tzinfo=tz.utc) if self.validTo else None
        return HierarchicalCode(
            code.id,
            code.name,
            code.description,
            code.valid_from,
            code.valid_to,
            vf,
            vt,
            codes,
        )


class JsonHierarchy(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a hierarchy."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    hierarchicalCodes: Sequence[JsonHierarchicalCode] = ()

    def to_model(self, codelists: Sequence[Codelist]) -> Hierarchy:
        """Converts a JsonHierarchy to a standard hierarchy."""
        return Hierarchy(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            codes=[i.to_model(codelists) for i in self.hierarchicalCodes],
        )


class JsonHierarchies(Struct, frozen=True):
    """SDMX-JSON payload for hierarchies."""

    codelists: Sequence[JsonCodelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()

    def to_model(self) -> Hierarchy:
        """Returns the requested hierarchy."""
        cls = [cl.to_model() for cl in self.codelists]
        return self.hierarchies[0].to_model(cls)


class JsonHierarchyAssociation(
    Struct, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for a hierarchy association."""

    id: str
    name: str
    agency: str
    linkedHierarchy: str
    linkedObject: str
    contextObject: str
    links: Sequence[JsonLink] = ()
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None

    def to_model(
        self,
        hierarchies: Sequence[JsonHierarchy],
        codelists: Sequence[JsonCodelist],
    ) -> HierarchyAssociation:
        """Converts a FusionHierarchyAssocation to a standard association."""
        cls = [cl.to_model() for cl in codelists]
        m = find_by_urn(hierarchies, self.linkedHierarchy).to_model(cls)
        lnk = list(
            filter(
                lambda i: hasattr(i, "rel") and i.rel == "UserDefinedOperator",
                self.links,
            )
        )
        return HierarchyAssociation(
            id=self.id,
            name=self.name,
            agency=self.agency,
            hierarchy=m,
            component_ref=self.linkedObject,
            context_ref=self.contextObject,
            description=self.description,
            version=self.version,
            operator=lnk[0].urn if lnk else None,
        )


class JsonHierarchyMessage(Struct, frozen=True):
    """SDMX-JSON payload for /hierarchy queries."""

    data: JsonHierarchies

    def to_model(self) -> Hierarchy:
        """Returns the requested hierarchy."""
        return self.data.to_model()


class JsonHierarchyAssociations(Struct, frozen=True):
    """SDMX-JSON payload for hierarchy associations."""

    codelists: Sequence[JsonCodelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()
    hierarchyassociations: Sequence[JsonHierarchyAssociation] = ()

    def to_model(self) -> Sequence[HierarchyAssociation]:
        """Returns the requested hierarchy associations."""
        return [
            ha.to_model(self.hierarchies, self.codelists)
            for ha in self.hierarchyassociations
        ]


class JsonHierarchyAssociationMessage(Struct, frozen=True):
    """SDMX-JSON payload for hierarchy associations messages."""

    data: JsonHierarchyAssociations

    def to_model(self) -> Sequence[HierarchyAssociation]:
        """Returns the requested hierarchy associations."""
        return self.data.to_model()



"""API for FMR readers."""

from dataclasses import dataclass
from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class Deserializer(Protocol):
    """Parses a message and return domain objects."""

    def to_model(self) -> Any:
        """Returns the domain objects."""


@dataclass
class Deserializers:
    """Collection of deserializers for a format."""

    agencies: Deserializer
    categories: Deserializer
    codes: Deserializer
    concepts: Deserializer
    dataflow: Deserializer
    providers: Deserializer
    schema: Deserializer
    hier_assoc: Deserializer
    hierarchy: Deserializer
    report: Deserializer
    mapping: Deserializer
    code_map: Deserializer



"""Collection of Fusion-JSON schemas for content constraints."""

from typing import Dict, Sequence

from msgspec import Struct


class FusionKeyValue(Struct, frozen=True):
    """Fusion-JSON payload for the list of allowed values per component."""

    values: Sequence[str]


class FusionContentConstraint(Struct, frozen=True):
    """Fusion-JSON payload for a content constraint."""

    includeCube: Dict[str, FusionKeyValue] = {}

    def to_map(self) -> Dict[str, Sequence[str]]:
        """Gets the list of allowed values for a component."""
        return {k: v.values for k, v in self.includeCube.items()}



ALLOWED_ERRORS_CONTENT = [
    "content type is empty",
    # QName value error on StructureSpecific xml files
    "of the xsi:type attribute does not resolve to a type definition.",
    "The type definition is abstract.",
]



"""Pandas SDMX Dataset."""

import pandas as pd

from pysdmx.model.dataset import Dataset


class PandasDataset(Dataset, frozen=False, kw_only=True):
    """Class related to Dataset, using Pandas Dataframe.

    It is based on SDMX Dataset and has Pandas Dataframe compatibility to
    withhold data.

    Args:
        attributes: Attributes at dataset level-
        data: Dataframe.
        structure:
        URN or Schema related to this Dataset
        (DSD, Dataflow, ProvisionAgreement)
        short_urn: Combination of Agency_id, Id and Version.
    """

    data: pd.DataFrame



"""Collection of SDMX-JSON schemas for common artefacts."""

from datetime import datetime
from typing import Optional, Sequence, Union

import msgspec

from pysdmx.model import ArrayBoundaries, Codelist, Facets
from pysdmx.util import find_by_urn


class JsonAnnotation(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for annotations."""

    id: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None
    value: Optional[str] = None
    text: Optional[str] = None


class JsonTextFormat(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for TextFormat."""

    dataType: str
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    minValue: Optional[Union[int, float]] = None
    maxValue: Optional[Union[int, float]] = None
    startValue: Optional[Union[int, float]] = None
    endValue: Optional[Union[int, float]] = None
    decimals: Optional[int] = None
    pattern: Optional[str] = None
    startTime: Optional[datetime] = None
    endTime: Optional[datetime] = None
    isSequence: bool = False
    isMultilingual: bool = False
    sentinelValues: Optional[Sequence[str]] = None
    timeInterval: Optional[str] = None
    interval: Optional[int] = None


class JsonRepresentation(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for core representation."""

    enumerationFormat: Optional[JsonTextFormat] = None
    enumeration: Optional[str] = None
    format: Optional[JsonTextFormat] = None
    minOccurs: Optional[int] = None
    maxOccurs: Optional[int] = None

    def to_facets(self) -> Optional[Facets]:
        """Return a Facets domain object."""
        if self.enumeration:
            fmt = self.enumerationFormat
        else:
            fmt = self.format
        if fmt and (
            fmt.minLength
            or fmt.maxLength
            or fmt.isSequence
            or fmt.minValue
            or fmt.maxValue
            or fmt.startValue
            or fmt.endValue
            or fmt.decimals
            or fmt.pattern
            or fmt.startTime
            or fmt.endTime
        ):
            return Facets(
                min_length=fmt.minLength,
                max_length=fmt.maxLength,
                is_sequence=fmt.isSequence,
                min_value=fmt.minValue,
                max_value=fmt.maxValue,
                start_value=fmt.startValue,
                end_value=fmt.endValue,
                decimals=fmt.decimals,
                pattern=fmt.pattern,
                start_time=fmt.startTime,
                end_time=fmt.endTime,
            )
        else:
            return None

    def to_enumeration(
        self,
        codelists: Sequence[Codelist],
        valid: Sequence[str],
    ) -> Optional[Codelist]:
        """Returns the list of codes allowed for this component."""
        if self.enumeration:
            a = find_by_urn(codelists, self.enumeration)
            codes = [c for c in a.codes if not valid or c.id in valid]
            return msgspec.structs.replace(a, items=codes)
        return None

    def to_array_def(self) -> Optional[ArrayBoundaries]:
        """Returns the array boundaries, if any."""
        if self.maxOccurs and self.maxOccurs > 1:
            m = self.minOccurs if self.minOccurs is not None else 0
            return ArrayBoundaries(m, self.maxOccurs)
        else:
            return None


class JsonLink(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for link objects."""

    urn: str
    rel: Optional[str] = None


class JsonHeader(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for message header."""

    links: Sequence[JsonLink] = ()



"""Dataset module."""

from datetime import date, datetime
from typing import Any, Dict, Optional, Union

from msgspec import Struct

from pysdmx.model import Schema
from pysdmx.model.message import ActionType


class SeriesInfo(Struct, frozen=True):
    """A group of related data, such as a time series, or a case series.

    Attributes:
        id: The unique ID for the series, such as a series key.
        name: The series's name (aka title), if any.
        obs_count: The number of observations available in the series.
        start_period: The oldest period for which data are available.
        end_period: The oldest period for which data are available.
        last_updated: When the series was last updated.
        is_active: Whether the series is still relevant or has been
            discontinued.
    """

    id: str
    name: Optional[str] = None
    obs_count: Optional[int] = None
    start_period: Optional[str] = None
    end_period: Optional[str] = None
    last_updated: Optional[datetime] = None
    is_active: bool = True

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__.keys():
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={v}")
        return ", ".join(out)


class Dataset(Struct, frozen=False, kw_only=True):
    """Core Dataset class."""

    structure: Union[str, Schema]
    attributes: Dict[str, Any] = {}
    action: ActionType = ActionType.Information
    reporting_begin: Optional[date] = None
    reporting_end: Optional[date] = None
    data_extraction_date: Optional[date] = None
    valid_from: Optional[date] = None
    valid_to: Optional[date] = None
    publication_year: Optional[date] = None
    publication_period: Optional[date] = None
    set_id: Optional[str] = None

    @property
    def short_urn(self) -> str:
        """Meaningful part of the URN.

        Returns:
            URN formatted string
        """
        if isinstance(self.structure, str):
            structure_type, unique_id = self.structure.split("=", maxsplit=1)
            structure = structure_type.rsplit(".", maxsplit=1)[1]
            return f"{structure}={unique_id}"
        else:
            s = self.structure
            return f"{s.context}={s.agency}:{s.id}({s.version})"



"""Utility functions for pysdmx query API."""

from parsy import ParseError  # type: ignore[import-untyped]

from pysdmx.api.dc.query._model import Filter
from pysdmx.api.dc.query._py_parser import py_parser
from pysdmx.api.dc.query._sql_parser import sql_parser
from pysdmx.errors import Invalid


def parse_query(query: str) -> Filter:
    """Parse a query string into a sequence of filters."""
    try:
        return py_parser.parse(query)
    except ParseError:
        try:
            return sql_parser.parse(query)
        except ParseError as pe:
            raise Invalid(
                "Unparseable query",
                (
                    "The query could not be parsed. "
                    "It must be a SQL WHERE clause or "
                    "a Python boolean expression. "
                    f"The query was: {query}"
                ),
            ) from pe


__all__ = ["parse_query"]



"""SDMX 2.0 CSV reader module."""

from io import StringIO
from typing import Dict

import pandas as pd

from pysdmx.errors import Invalid
from pysdmx.io.pd import PandasDataset
from pysdmx.model.message import ActionType

ACTION_SDMX_CSV_MAPPER_READING = {
    "A": ActionType.Append,
    "D": ActionType.Delete,
    "R": ActionType.Replace,
    "I": ActionType.Information,
}


def __generate_dataset_from_sdmx_csv(data: pd.DataFrame) -> PandasDataset:
    # Extract Structure type and structure id
    action = ActionType.Information
    if "ACTION" in data.columns:
        unique_values = list(data["ACTION"].unique())
        if len(unique_values) > 1 and "D" in unique_values:
            unique_values.remove("D")
            data = data[data["ACTION"] != "D"]
        if len(unique_values) == 1:  # If there is only one value, use it
            action_value = unique_values[0]
            if action_value not in ACTION_SDMX_CSV_MAPPER_READING:
                raise Invalid(
                    "Invalid value on ACTION column",
                    "Invalid SDMX-CSV 2.0 file. "
                    "Check the docs for the proper values on ACTION column.",
                )
            action = ACTION_SDMX_CSV_MAPPER_READING[action_value]
        else:
            raise Invalid(
                "Invalid value on ACTION column",
                "Invalid SDMX-CSV 2.0 file. "
                "Cannot have more than one value on ACTION column, "
                "or 2 if D is present",
            )
    # For SDMX-CSV version 2, use 'STRUCTURE_ID'
    # column as the structure id and 'STRUCTURE' as the structure type
    structure_id = data["STRUCTURE_ID"].iloc[0]
    structure_type = data["STRUCTURE"].iloc[0]
    # Drop 'STRUCTURE' and 'STRUCTURE_ID' columns from DataFrame
    df_csv = data.drop(["STRUCTURE", "STRUCTURE_ID"], axis=1)

    if structure_type == "DataStructure".lower():
        urn = (
            "urn:sdmx:org.sdmx.infomodel.datastructure."
            f"DataStructure={structure_id}"
        )
    elif structure_type == "DataFlow".lower():
        urn = (
            "urn:sdmx:org.sdmx.infomodel.datastructure."
            f"DataFlow={structure_id}"
        )
    elif structure_type == "dataprovision":
        urn = (
            f"urn:sdmx:org.sdmx.infomodel.registry."
            f"ProvisionAgreement={structure_id}"
        )
    else:
        raise Invalid(
            "Invalid value on STRUCTURE column",
            "Invalid SDMX-CSV 2.0 file. "
            "Check the docs for the proper values on STRUCTURE column.",
        )
    # Extract dataset attributes from sdmx-csv (all values are the same)
    attributes = {
        col: df_csv[col].iloc[0]
        for col in df_csv.columns
        if df_csv[col].nunique() == 1
    }
    for col in attributes:
        df_csv = df_csv.drop(col, axis=1)

    # Return a Dataset object with the extracted information
    return PandasDataset(
        structure=urn,
        data=df_csv,
        attributes=attributes,
        action=action,
    )


def read(infile: str) -> Dict[str, PandasDataset]:
    """Reads csv file and returns a payload dictionary.

    Args:
        infile: Path to file, str.

    Returns:
        payload: dict.

    Raises:
        Invalid: If it is an invalid CSV file.
    """
    # Get Dataframe from CSV file
    df_csv = pd.read_csv(StringIO(infile))
    # Drop empty columns
    df_csv = df_csv.dropna(axis=1, how="all")

    # Determine SDMX-CSV version based on column names
    if (
        "STRUCTURE" not in df_csv.columns
        or "STRUCTURE_ID" not in df_csv.columns
    ):
        # Raise an exception if the CSV file is not in SDMX-CSV format
        raise Invalid(
            "Only SDMX-CSV 2.0 is allowed",
            "Invalid SDMX-CSV 2.0 file. "
            "Check the docs for the proper structure on content.",
        )

    # Convert all columns to strings
    df_csv = df_csv.astype("str")
    # Check if any column headers contain ':', indicating mode, label or text
    mode_label_text = any(":" in x for x in df_csv.columns)

    id_column = "STRUCTURE_ID"
    # If mode, label or text is present, modify the DataFrame
    if mode_label_text:
        # Split the ID column to remove mode, label or text
        df_csv[id_column] = df_csv[id_column].map(lambda x: x.split(": ")[0])
        # Split the other columns to remove mode, label, or text
        sequence = 3
        for x in df_csv.columns[sequence:]:
            df_csv[x.split(":")[0]] = df_csv[x].map(
                lambda x: x.split(": ", 2)[0], na_action="ignore"
            )
            # Delete the original columns
            del df_csv[x]

    # Grouping columns to separate datasets
    grouping_columns = ["STRUCTURE", "STRUCTURE_ID"]
    # Separate SDMX-CSV in different datasets per Structure ID
    list_df = [data for _, data in df_csv.groupby(grouping_columns)]

    # Create a payload dictionary to store datasets with the
    # different unique_ids as keys
    payload = {}
    for df in list_df:
        # Generate a dataset from each subset of the DataFrame
        dataset = __generate_dataset_from_sdmx_csv(data=df)

        # Add the dataset to the payload dictionary
        payload[dataset.short_urn] = dataset

    # Return the payload generated
    return payload



"""Collection of Fusion-JSON schemas for reference metadata queries."""

from typing import Sequence

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.model.metadata import (
    merge_attributes,
    MetadataAttribute,
    MetadataReport,
)


class FusionMetadataReport(Struct, frozen=True):
    """Fusion-JSON payload for a metadata report."""

    id: str
    names: Sequence[FusionString]
    metadataflow: str
    targets: Sequence[str]
    attributes: Sequence[MetadataAttribute]


class FusionMetadataSets(Struct, frozen=True):
    """Fusion-JSON payload for the list of metadata sets."""

    metadatasets: Sequence[FusionMetadataReport]


class FusionMetadataMessage(Struct, frozen=True):
    """Fusion-JSON payload for /metadata queries."""

    data: FusionMetadataSets

    def __create_report(self, r: FusionMetadataReport) -> MetadataReport:
        attrs = merge_attributes(r.attributes)
        return MetadataReport(
            r.id, r.names[0].value, r.metadataflow, r.targets, attrs
        )

    def to_model(self, fetch_all: bool = False) -> Sequence[MetadataReport]:
        """Returns the requested metadata report(s)."""
        if fetch_all:
            return [self.__create_report(r) for r in self.data.metadatasets]
        else:
            return [self.__create_report(self.data.metadatasets[0])]



"""Collection of Fusion-JSON schemas for codes and codelists."""

from datetime import datetime, timedelta, timezone as tz
from typing import Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import (
    FusionAnnotation,
    FusionLink,
    FusionString,
)
from pysdmx.model import (
    Code,
    Codelist as CL,
    HierarchicalCode,
    Hierarchy as HCL,
    HierarchyAssociation as HA,
)
from pysdmx.util import find_by_urn, parse_item_urn


class FusionCode(Struct, frozen=True):
    """Fusion-JSON payload for codes."""

    id: str
    annotations: Sequence[FusionAnnotation] = ()
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()

    def __handle_date(self, datestr: str) -> datetime:
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S%z")

    def __get_val(
        self, a: FusionAnnotation
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        vals = a.title.split("/")
        if a.title.startswith("/"):
            return (None, self.__handle_date(vals[1]))
        else:
            valid_from = self.__handle_date(vals[0])
            valid_to = self.__handle_date(vals[1]) if vals[1] else None
            return (valid_from, valid_to)

    def to_model(self) -> Code:
        """Converts a FusionCode to a standard code."""
        vp = [a for a in self.annotations if a.type == "FR_VALIDITY_PERIOD"]
        vf, vt = self.__get_val(vp[0]) if vp else (None, None)
        return Code(
            id=self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            valid_from=vf,
            valid_to=vt,
        )


class FusionCodelist(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a codelist."""

    id: str
    urn: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    items: Sequence[FusionCode] = ()

    def to_model(self) -> CL:
        """Converts a JsonCodelist to a standard codelist."""
        t = "codelist" if "Codelist" in self.urn else "valuelist"
        return CL(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            items=[i.to_model() for i in self.items],
            sdmx_type=t,  # type: ignore[arg-type]
        )


class FusionCodelistMessage(Struct, frozen=True):
    """Fusion-JSON payload for /codelist queries."""

    Codelist: Sequence[FusionCodelist] = ()
    ValueList: Sequence[FusionCodelist] = ()

    def to_model(self) -> CL:
        """Returns the requested codelist."""
        if self.Codelist:
            return self.Codelist[0].to_model()
        else:
            return self.ValueList[0].to_model()


class FusionHierarchicalCode(Struct, frozen=True):
    """Fusion-JSON payload for hierarchical codes."""

    code: str
    validFrom: Optional[int] = None
    validTo: Optional[int] = None
    codes: Sequence["FusionHierarchicalCode"] = ()

    def __find_code(self, codelists: Sequence[CL], urn: str) -> Code:
        r = parse_item_urn(urn)
        f = [
            c
            for c in codelists
            if (
                c.agency == r.agency
                and c.id == r.id
                and c.version == r.version
            )
        ]
        return [c for c in f[0].codes if c.id == r.item_id][0]

    def __convert_epoch(self, epoch: int) -> datetime:
        if epoch < 0:
            return datetime(1970, 1, 1, tzinfo=tz.utc) + timedelta(
                milliseconds=epoch
            )
        else:
            return datetime.fromtimestamp(epoch / 1000, tz.utc)

    def to_model(self, codelists: Sequence[CL]) -> HierarchicalCode:
        """Converts a FusionHierarchicalCode to a hierachical code."""
        code = self.__find_code(codelists, self.code)
        rvf = self.__convert_epoch(self.validFrom) if self.validFrom else None
        rvt = self.__convert_epoch(self.validTo) if self.validTo else None
        codes = [c.to_model(codelists) for c in self.codes]
        return HierarchicalCode(
            code.id,
            code.name,
            code.description,
            code.valid_from,
            code.valid_to,
            rvf,
            rvt,
            codes,
        )


class FusionHierarchy(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a hierarchy."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    codes: Sequence[FusionHierarchicalCode] = ()

    def to_model(self, codelists: Sequence[CL]) -> HCL:
        """Converts a FusionHierarchy to a standard hierarchy."""
        return HCL(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            codes=[i.to_model(codelists) for i in self.codes],
        )


class FusionHierarchyAssociation(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for a hierarchy association."""

    id: str
    names: Sequence[FusionString]
    agency: str
    hierarchyRef: str
    linkedStructureRef: str
    contextRef: str
    links: Sequence[FusionLink] = ()
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"

    def to_model(
        self,
        hierarchies: Sequence[FusionHierarchy],
        codelists: Sequence[FusionCodelist],
    ) -> HA:
        """Converts a FusionHierarchyAssocation to a standard association."""
        cls = [cl.to_model() for cl in codelists]
        m = find_by_urn(hierarchies, self.hierarchyRef).to_model(cls)
        return HA(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            hierarchy=m,
            component_ref=self.linkedStructureRef,
            context_ref=self.contextRef,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            operator=self.links[0].urn if self.links else None,
        )


class FusionHierarchyMessage(Struct, frozen=True):
    """Fusion-JSON payload for /hierarchy queries."""

    Codelist: Sequence[FusionCodelist] = ()
    Hierarchy: Sequence[FusionHierarchy] = ()

    def to_model(self) -> HCL:
        """Returns the requested hierarchy."""
        cls = [cl.to_model() for cl in self.Codelist]
        return self.Hierarchy[0].to_model(cls)


class FusionHierarchyAssociationMessage(Struct, frozen=True):
    """Fusion-JSON payload for hierarchy associations."""

    Codelist: Sequence[FusionCodelist] = ()
    Hierarchy: Sequence[FusionHierarchy] = ()
    HierarchyAssociation: Sequence[FusionHierarchyAssociation] = ()

    def to_model(self) -> Sequence[HA]:
        """Returns the requested hierarchy associations."""
        return [
            ha.to_model(self.Hierarchy, self.Codelist)
            for ha in self.HierarchyAssociation
        ]
